# coding: utf-8
###########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

"""contain the AxisProcess
"""

__authors__ = ["C.Nemoz", "H.Payno"]
__license__ = "MIT"
__date__ = "19/03/2019"

import logging
from typing import Optional, Union

import numpy
from nabu.estimation.cor import (
    CenterOfRotation,
    CenterOfRotationAdaptiveSearch,
    CenterOfRotationGrowingWindow,
    CenterOfRotationSlidingWindow,
    CenterOfRotationOctaveAccurate,
)
from nabu.pipeline.estimators import SinoCORFinder, CORFinder
from nabu.resources.nxflatfield import update_dataset_info_flats_darks
from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess
from tomwer.core.utils.deprecation import deprecated_warning

import tomwer.version
from tomwer.core.process.reconstruction.utils.cor import absolute_pos_to_relative
from tomwer.core.process.task import Task
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils import image, logconfig
from tomwer.core.utils.scanutils import data_identifier_to_scan
from tomwer.utils import docstring

from .mode import AxisMode
from .params import (
    DEFAULT_CMP_N_SUBSAMPLING_Y,
    DEFAULT_CMP_NEAR_POS,
    DEFAULT_CMP_NEAR_WIDTH,
    DEFAULT_CMP_OVERSAMPLING,
    DEFAULT_CMP_TAKE_LOG,
    DEFAULT_CMP_THETA,
    AxisRP,
)
from .projectiontype import ProjectionType

try:
    from nabu.pipeline.estimators import CompositeCOREstimator
except ImportError:
    has_composite_cor_finder = False
else:
    has_composite_cor_finder = True
from silx.io.url import DataUrl
from silx.io.utils import h5py_read_dataset
from silx.io.utils import open as open_hdf5

_logger = logging.getLogger(__name__)
if not has_composite_cor_finder:
    _logger.warning("No composite cor finder found at nabu level")
# vertically, work on a window having only a percentage of the frame.
pc_height = 10.0 / 100.0
# horizontally. Global method supposes the COR is more or less in center
# % of the detector:

pc_width = 50.0 / 100.0


def _absolute_pos_to_relative_with_warning(
    absolute_pos: float, det_width: Optional[int]
):
    """
    nabu returns the value as absolute. tomwer needs it as relative
    Also handle the case (unlikely) the detector width cannot be found
    """
    if det_width is None:
        det_width = 2048
        _logger.warning("unable to find image width. Set width to 2048")
    else:
        det_width = det_width
    return absolute_pos_to_relative(absolute_pos=absolute_pos, det_width=det_width)


def adapt_tomwer_scan_to_nabu(scan: TomwerScanBase):
    """simple util to convert tomwer scan to a nabu DataAnalizer and
    updating infos regarding flat and dark if needed
    """
    dataset_infos = scan.to_nabu_dataset_analyser()
    if isinstance(scan, NXtomoScan):
        try:
            update_dataset_info_flats_darks(
                dataset_infos,
                flatfield_mode=None,
            )
        except ValueError as exception:
            # nabu raise an error if no darks / flats set. But this can make sense at this stage if the NXtomo has no
            # raw dark / flat and is already normalized. In this case only fire a warning
            if (
                scan.reduced_darks is not None
                and len(scan.reduced_darks) > 0
                and scan.reduced_flats is not None
                and len(scan.reduced_flats) > 0
            ):
                raise exception
            else:
                _logger.warning(
                    "Fail to update nabu dataset info flats and darks. Expected if the dataset contains already normalized projections"
                )

    return dataset_infos


def compute_cor_nabu_growing_window(
    radio_1: numpy.ndarray,
    radio_2: numpy.ndarray,
    side: str,
    padding_mode,
    flip_frame_2_lr=True,
    horz_fft_width=False,
):
    """
    Call nabu.preproc.alignement.CenterOfRotationGrowingWindow.find_shift

    :param radio_1:
    :param radio_2:
    :param padding_mode: padding mode
    :param str side: side of the cor
    :param bool flip_frame_2_lr: if True will left-right flip the second frame
    :param horz_fft_width:

    :return:
    """
    nabu_class = CenterOfRotationGrowingWindow(horz_fft_width=horz_fft_width)
    # value return is relative
    res = nabu_class.find_shift(
        img_1=radio_1,
        img_2=numpy.fliplr(radio_2) if flip_frame_2_lr else radio_2,
        side=side,
        roi_yxhw=None,
        padding_mode=padding_mode,
        median_filt_shape=None,
    )
    if isinstance(res, numpy.ndarray):
        if len(res) == 1:
            return res[0]
        else:
            raise ValueError(
                "nabu rsult is expected to be a scalar, numpy array found. Please upgrade nabu this issue is expected to be solved"
            )
    else:
        return res


def compute_cor_nabu_growing_window_radios(
    scan: TomwerScanBase,
):
    """
    Call nabu.preproc.alignement.CenterOfRotationGrowingWindow.find_shift

    :param TomwerScanBase scan:

    :return:
    """
    has_darks = scan.reduced_darks is not None and len(scan.reduced_darks) > 0
    has_flats = scan.reduced_flats is not None and len(scan.reduced_flats) > 0

    projection_angles = scan.get_proj_angle_url()
    projection_angles_i = {
        value.path(): key for key, value in projection_angles.items()
    }
    url_radio_1, url_radio_2 = AxisTask.get_inputs_urls(scan=scan)
    angle_radio_1 = float(projection_angles_i[url_radio_1.url.path()])
    angle_radio_2 = float(projection_angles_i[url_radio_2.url.path()])
    radio_angles = tuple(numpy.radians((angle_radio_1, angle_radio_2)))

    corfinder = CORFinder(
        dataset_info=adapt_tomwer_scan_to_nabu(scan),
        method="growing-window",
        do_flatfield=has_darks and has_flats,
        cor_options=scan.axis_params.get_nabu_cor_options_as_dict(),
        radio_angles=radio_angles,
        logger=_logger,
    )
    res = corfinder.find_cor()  # Returns absolute cor
    if isinstance(res, numpy.ndarray):
        if len(res) == 1:
            res = res[0]
        else:
            raise ValueError(
                "nabu rsult is expected to be a scalar, numpy array found. Please upgrade nabu this issue is expected to be solved"
            )

    return _absolute_pos_to_relative_with_warning(
        absolute_pos=res, det_width=scan.dim_1
    )


def compute_cor_nabu_growing_window_sinogram(
    scan: TomwerScanBase,
):
    """
    Call nabu.preproc.alignement.CenterOfRotationGrowingWindow.find_shift

    :param TomwerScanBase scan:

    :return:
    """
    has_darks = scan.reduced_darks is not None and len(scan.reduced_darks) > 0
    has_flats = scan.reduced_flats is not None and len(scan.reduced_flats) > 0

    corfinder = SinoCORFinder(
        dataset_info=adapt_tomwer_scan_to_nabu(scan),
        method="sino-growing-window",
        slice_idx=scan.axis_params.sinogram_line or "middle",
        subsampling=scan.axis_params.sinogram_subsampling,
        do_flatfield=has_darks and has_flats,
        cor_options=scan.axis_params.get_nabu_cor_options_as_dict(),
        logger=_logger,
    )
    res = corfinder.find_cor()
    if isinstance(res, numpy.ndarray):
        if len(res) == 1:
            res = res[0]
        else:
            raise ValueError(
                "nabu rsult is expected to be a scalar, numpy array found. Please upgrade nabu this issue is expected to be solved"
            )

    return _absolute_pos_to_relative_with_warning(
        absolute_pos=res, det_width=scan.dim_1
    )


def compute_scan_sino_coarse_to_fine(scan):
    """
    Compute center of rotation from `sino-coarse-to-fine` algorithm for given
    scan
    :param scan:
    :return:
    """
    has_darks = scan.reduced_darks is not None and len(scan.reduced_darks) > 0
    has_flats = scan.reduced_flats is not None and len(scan.reduced_flats) > 0

    corfinder = SinoCORFinder(
        dataset_info=adapt_tomwer_scan_to_nabu(scan),
        method=AxisMode.sino_coarse_to_fine.value,
        slice_idx=scan.axis_params.sinogram_line or "middle",
        subsampling=scan.axis_params.sinogram_subsampling,
        do_flatfield=has_darks and has_flats,
        cor_options=scan.axis_params.get_nabu_cor_options_as_dict(),
        logger=_logger,
    )
    res = corfinder.find_cor()
    return _absolute_pos_to_relative_with_warning(
        absolute_pos=res, det_width=scan.dim_1
    )


def compute_scan_composite_coarse_to_fine(scan: TomwerScanBase):
    """
    Compute center of rotation from `sino-coarse-to-fine` algorithm for given
    scan
    :param scan:
    :return:
    """
    if not has_composite_cor_finder:
        _logger.error("unable to find nabu CompositeCOREstimator")
        return None

    (
        theta,
        n_subsampling_y,
        oversampling,
        take_log,
        near_pos,
        near_width,
    ) = get_composite_options(scan)

    # as the new corfinder is not yet merged in the main branch
    # allow some tolerance for the "side" argument that is there only
    # in the new one

    cor_options = scan.axis_params.get_nabu_cor_options_as_dict()
    for key in "low_pass", "high_pass":
        if key in cor_options:
            cor_options[key] = int(cor_options[key])
    corfinder = CompositeCOREstimator(
        dataset_info=adapt_tomwer_scan_to_nabu(scan),
        theta_interval=theta,
        n_subsampling_y=n_subsampling_y,
        oversampling=oversampling,
        cor_options=cor_options,
        logger=_logger,
        take_log=take_log,
    )
    res = corfinder.find_cor()
    return _absolute_pos_to_relative_with_warning(
        absolute_pos=res, det_width=scan.dim_1
    )


def get_composite_options(scan):
    theta = scan.axis_params.composite_options.get("theta", DEFAULT_CMP_THETA)
    n_subsampling_y = scan.axis_params.composite_options.get(
        "n_subsampling_y", DEFAULT_CMP_N_SUBSAMPLING_Y
    )
    oversampling = scan.axis_params.composite_options.get(
        "oversampling", DEFAULT_CMP_OVERSAMPLING
    )
    take_log = scan.axis_params.composite_options.get("take_log", DEFAULT_CMP_TAKE_LOG)

    near_pos = scan.axis_params.composite_options.get("near_pos", DEFAULT_CMP_NEAR_POS)
    near_width = scan.axis_params.composite_options.get(
        "near_width", DEFAULT_CMP_NEAR_WIDTH
    )
    return theta, n_subsampling_y, oversampling, take_log, near_pos, near_width


def compute_scan_cor_nabu_growing_window(scan):
    """
    Call to nabu.preproc.alignment.CenterOfRotation from the scan axis_params
    value.

    :param `.TomoBase` scan: scan to process
    :return: Union[float, None]
    """
    if not scan.axis_params.use_sinogram:
        radio_1, radio_2 = AxisTask.get_inputs(scan=scan)
        if radio_1 is None or radio_2 is None:
            raise NoAxisUrl("Unable to find projections for nabu axis calculation")
    else:
        radio_1, radio_2 = None, None

    _logger.info(
        "compute scan axis from nabu CenterOfRotationGrowingWindow with padding "
        "mode {} and side {}. Use sinogram: {}".format(
            scan.axis_params.padding_mode,
            scan.axis_params.side,
            scan.axis_params.use_sinogram,
        )
    )

    if scan.axis_params.use_sinogram:
        return compute_cor_nabu_growing_window_sinogram(scan=scan)
    else:
        return compute_cor_nabu_growing_window_radios(scan=scan)


def compute_cor_nabu_sliding_window(
    radio_1: numpy.ndarray,
    radio_2: numpy.ndarray,
    side: str,
    padding_mode,
    flip_frame_2_lr=True,
    horz_fft_width=False,
):
    """
    Call nabu.preproc.alignement.CenterOfRotationSlidingWindow.find_shift

    :param radio_1:
    :param radio_2:
    :param padding_mode:
    :param str side: side of the cor
    :param horz_fft_width:
    :param bool flip_frame_2_lr: if True will left-right flip the second frame
    :param half_acq_cor_guess: The approximate position of the rotation axis
                               from the image center. Optional. When given a
                               special algorithm is used which can work also
                               in half-tomo conditions.

    :return:
    """
    nabu_class = CenterOfRotationSlidingWindow(horz_fft_width=horz_fft_width)
    res = nabu_class.find_shift(
        img_1=radio_1,
        img_2=numpy.fliplr(radio_2) if flip_frame_2_lr else radio_2,
        side=side,
        roi_yxhw=None,
        padding_mode=padding_mode,
        median_filt_shape=None,
    )
    return res


def compute_cor_nabu_sliding_window_radios(
    scan,
):
    """
    Call nabu.preproc.alignement.CenterOfRotationGrowingWindow.find_shift

    :param TomwerScanBase scan:

    :return:
    """
    has_darks = scan.reduced_darks is not None and len(scan.reduced_darks) > 0
    has_flats = scan.reduced_flats is not None and len(scan.reduced_flats) > 0

    projection_angles = scan.get_proj_angle_url()
    projection_angles_i = {
        value.path(): key for key, value in projection_angles.items()
    }
    url_radio_1, url_radio_2 = AxisTask.get_inputs_urls(scan=scan)
    angle_radio_1 = float(projection_angles_i[url_radio_1.url.path()])
    angle_radio_2 = float(projection_angles_i[url_radio_2.url.path()])
    radio_angles = tuple(numpy.radians((angle_radio_1, angle_radio_2)))

    corfinder = CORFinder(
        dataset_info=adapt_tomwer_scan_to_nabu(scan),
        method="sliding-window",
        do_flatfield=has_darks and has_flats,
        cor_options=scan.axis_params.get_nabu_cor_options_as_dict(),
        radio_angles=radio_angles,
        logger=_logger,
    )
    res = corfinder.find_cor()  # Returns absolute cor.
    return _absolute_pos_to_relative_with_warning(
        absolute_pos=res, det_width=scan.dim_1
    )


def compute_cor_nabu_sliding_window_sinogram(
    scan,
):
    """
    Call nabu.preproc.alignement.CenterOfRotationGrowingWindow.find_shift

    :param TomwerScanBase scan:

    :return:
    """
    has_darks = scan.reduced_darks is not None and len(scan.reduced_darks) > 0
    has_flats = scan.reduced_flats is not None and len(scan.reduced_flats) > 0

    corfinder = SinoCORFinder(
        dataset_info=adapt_tomwer_scan_to_nabu(scan),
        method="sino-sliding-window",
        slice_idx=scan.axis_params.sinogram_line or "middle",
        subsampling=scan.axis_params.sinogram_subsampling,
        do_flatfield=has_darks and has_flats,
        cor_options=scan.axis_params.get_nabu_cor_options_as_dict(),
        logger=_logger,
    )
    res = corfinder.find_cor()
    return _absolute_pos_to_relative_with_warning(
        absolute_pos=res, det_width=scan.dim_1
    )


def compute_scan_cor_nabu_sliding_window(scan):
    """
    Call to nabu.preproc.alignment.CenterOfRotation from the scan axis_params
    value.

    :param `.TomoBase` scan: scan to process
    :return: Union[float, None]
    """
    if not scan.axis_params.use_sinogram:
        radio_1, radio_2 = AxisTask.get_inputs(scan=scan)
        if radio_1 is None or radio_2 is None:
            raise NoAxisUrl("Unable to find projections for nabu axis calculation")
    else:
        radio_1 = radio_2 = None

    _logger.info(
        "compute scan axis from nabu CenterOfRotationSlidingWindow with padding "
        "mode {} and side {}. Use sinogram: {}".format(
            scan.axis_params.padding_mode,
            scan.axis_params.side,
            scan.axis_params.use_sinogram,
        )
    )

    if scan.axis_params.use_sinogram:
        return compute_cor_nabu_sliding_window_sinogram(scan=scan)
    else:
        return compute_cor_nabu_sliding_window_radios(scan=scan)


def compute_scan_fourier_angles(scan):
    """
    run 'scan_fourier_angles' algorithm for the requested scan
    """
    has_darks = scan.reduced_darks is not None and len(scan.reduced_darks) > 0
    has_flats = scan.reduced_flats is not None and len(scan.reduced_flats) > 0

    corfinder = SinoCORFinder(
        dataset_info=adapt_tomwer_scan_to_nabu(scan),
        method="fourier-angles",
        slice_idx=scan.axis_params.sinogram_line or "middle",
        subsampling=scan.axis_params.sinogram_subsampling,
        do_flatfield=has_darks and has_flats,
        cor_options=scan.axis_params.get_nabu_cor_options_as_dict(),
        logger=_logger,
    )
    res = corfinder.find_cor()
    return _absolute_pos_to_relative_with_warning(
        absolute_pos=res, det_width=scan.dim_1
    )


def compute_scan_octave_accurate_radios(
    scan,
):
    """
    Call nabu.preproc.alignement.CenterOfRotationGrowingWindow.find_shift

    :param TomwerScanBase scan:

    :return:
    """
    has_darks = scan.reduced_darks is not None and len(scan.reduced_darks) > 0
    has_flats = scan.reduced_flats is not None and len(scan.reduced_flats) > 0

    has_darks = scan.reduced_darks is not None and len(scan.reduced_darks) > 0
    has_flats = scan.reduced_flats is not None and len(scan.reduced_flats) > 0

    projection_angles = scan.get_proj_angle_url()
    projection_angles_i = {
        value.path(): key for key, value in projection_angles.items()
    }
    url_radio_1, url_radio_2 = AxisTask.get_inputs_urls(scan=scan)
    angle_radio_1 = float(projection_angles_i[url_radio_1.url.path()])
    angle_radio_2 = float(projection_angles_i[url_radio_2.url.path()])
    radio_angles = tuple(numpy.radians((angle_radio_1, angle_radio_2)))

    corfinder = CORFinder(
        dataset_info=adapt_tomwer_scan_to_nabu(scan),
        method="octave-accurate",
        do_flatfield=has_darks and has_flats,
        cor_options=scan.axis_params.get_nabu_cor_options_as_dict(),
        radio_angles=radio_angles,
        logger=_logger,
    )
    res = corfinder.find_cor()
    return _absolute_pos_to_relative_with_warning(
        absolute_pos=res, det_width=scan.dim_1
    )


def compute_scan_octave_accurate(scan):
    """
    Compute center of rotation from `octave-accurate` algorithm
    scan
    :param scan:
    :return:
    """
    cor_options = scan.axis_params.get_nabu_cor_options_as_dict()
    radio_1, radio_2 = AxisTask.get_inputs(scan=scan)
    extra_options = {}
    for key in "low_pass", "high_pass":
        if key in cor_options:
            extra_options[key] = float(cor_options[key])

    corfinder = CenterOfRotationOctaveAccurate(cor_options=cor_options)
    res = corfinder.find_shift(
        img_1=radio_1,
        img_2=numpy.fliplr(radio_2) if scan.axis_params.flip_lr else radio_2,
        side=scan.axis_params.side,
        padding_mode=scan.axis_params.padding_mode,
        **extra_options,
    )
    return res


def compute_cor_nabu_centered_radios(
    scan,
):
    """
    Call nabu.preproc.alignement.CenterOfRotationGrowingWindow.find_shift

    :param TomwerScanBase scan:

    :return:
    """
    has_darks = scan.reduced_darks is not None and len(scan.reduced_darks) > 0
    has_flats = scan.reduced_flats is not None and len(scan.reduced_flats) > 0

    projection_angles = scan.get_proj_angle_url()
    projection_angles_i = {
        value.path(): key for key, value in projection_angles.items()
    }
    url_radio_1, url_radio_2 = AxisTask.get_inputs_urls(scan=scan)
    angle_radio_1 = float(projection_angles_i[url_radio_1.url.path()])
    angle_radio_2 = float(projection_angles_i[url_radio_2.url.path()])
    radio_angles = tuple(numpy.radians((angle_radio_1, angle_radio_2)))

    corfinder = CORFinder(
        dataset_info=adapt_tomwer_scan_to_nabu(scan),
        method="centered",
        do_flatfield=has_darks and has_flats,
        cor_options=scan.axis_params.get_nabu_cor_options_as_dict(),
        radio_angles=radio_angles,
        logger=_logger,
    )
    res = corfinder.find_cor()
    return _absolute_pos_to_relative_with_warning(
        absolute_pos=res, det_width=scan.dim_1
    )


def compute_cor_nabu_centered(
    radio_1: numpy.ndarray,
    radio_2: numpy.ndarray,
    padding_mode,
    flip_frame_2_lr=True,
    horz_fft_width=False,
    vert_fft_width=False,
):
    """
    Call nabu.preproc.alignement.CenterOfRotation.find_shift

    :param radio_1:
    :param radio_2:
    :param padding_mode:
    :param horz_fft_width:
    :param bool flip_frame_2_lr: if True will left-right flip the second frame
    :param half_acq_cor_guess: The approximate position of the rotation axis
                               from the image center. Optional. When given a
                               special algorithm is used which can work also
                               in half-tomo conditions.

    :return:
    """
    nabu_class = CenterOfRotation(
        horz_fft_width=horz_fft_width, vert_fft_width=vert_fft_width
    )
    return nabu_class.find_shift(
        img_1=radio_1,
        img_2=numpy.fliplr(radio_2) if flip_frame_2_lr else radio_2,
        roi_yxhw=None,
        padding_mode=padding_mode,
        median_filt_shape=None,
    )


def compute_scan_cor_nabu_centered(scan):
    """
    Call to nabu.preproc.alignment.CenterOfRotation from the scan axis_params
    value.

    :param `.TomoBase` scan: scan to process
    :return: Union[float, None]

    """
    assert scan.axis_params is not None
    radio_1, radio_2 = AxisTask.get_inputs(scan=scan)
    if radio_1 is None or radio_2 is None:
        raise NoAxisUrl("Unable to find projections for nabu axis calculation")

    _logger.info(
        "compute scan axis from nabu CenterOfRotation with padding "
        "mode %s" % scan.axis_params.padding_mode
    )

    return compute_cor_nabu_centered_radios(scan)


def compute_cor_nabu_global(
    radio_1: numpy.ndarray,
    radio_2: numpy.ndarray,
    padding_mode,
    flip_frame_2_lr=True,
    horz_fft_width=False,
):
    """
    Call nabu.preproc.alignement.CenterOfRotation.find_shift

    :param radio_1:
    :param radio_2:
    :param padding_mode:
    :param horz_fft_width:
    :param bool flip_frame_2_lr: if True will left-right flip the second frame
    :return:
    """
    nabu_class = CenterOfRotationAdaptiveSearch(horz_fft_width=horz_fft_width)
    return nabu_class.find_shift(
        img_1=radio_1,
        img_2=numpy.fliplr(radio_2) if flip_frame_2_lr else radio_2,
        roi_yxhw=None,
        padding_mode=padding_mode,
        median_filt_shape=None,
    )


def compute_cor_nabu_global_radios(
    scan: TomwerScanBase,
):
    """
    Call nabu.preproc.alignement.CenterOfRotationGrowingWindow.find_shift

    :param TomwerScanBase scan:

    :return:
    """
    has_darks = scan.reduced_darks is not None and len(scan.reduced_darks) > 0
    has_flats = scan.reduced_flats is not None and len(scan.reduced_flats) > 0

    projection_angles = scan.get_proj_angle_url()
    projection_angles_i = {
        value.path(): key for key, value in projection_angles.items()
    }
    url_radio_1, url_radio_2 = AxisTask.get_inputs_urls(scan=scan)
    angle_radio_1 = float(projection_angles_i[url_radio_1.url.path()])
    angle_radio_2 = float(projection_angles_i[url_radio_2.url.path()])
    radio_angles = tuple(numpy.radians((angle_radio_1, angle_radio_2)))

    corfinder = CORFinder(
        dataset_info=adapt_tomwer_scan_to_nabu(scan),
        method="global",
        do_flatfield=has_darks and has_flats,
        cor_options=scan.axis_params.get_nabu_cor_options_as_dict(),
        radio_angles=radio_angles,
        logger=_logger,
    )
    res = corfinder.find_cor()
    return _absolute_pos_to_relative_with_warning(
        absolute_pos=res, det_width=scan.dim_1
    )


def compute_scan_cor_nabu_global(scan):
    """
    Call to nabu.preproc.alignment.CenterOfRotation from the scan axis_params
    value.

    :param `.TomoBase` scan: scan to process
    :return: Union[float, None]

    """
    assert scan.axis_params is not None
    radio_1, radio_2 = AxisTask.get_inputs(scan=scan)
    if radio_1 is None or radio_2 is None:
        raise NoAxisUrl("Unable to find projections for nabu axis calculation")

    _logger.info(
        "compute scan axis from nabu CenterOfRotation with padding "
        "mode %s" % scan.axis_params.padding_mode
    )
    return compute_cor_nabu_global_radios(scan)


def get_stdmax_column(x: numpy.ndarray) -> float:
    """

    :param x:
    :return: column index of the maximal standard deviation
    """
    kernel_size = 5
    length = len(x)
    r = range(length - kernel_size)
    y = numpy.empty(length - kernel_size)
    for i in r:
        s = numpy.std(x[i : i + kernel_size])
        y[i] = s

    return y.argmax()


class NoAxisUrl(Exception):
    pass


class AxisTask(
    Task,
    SuperviseProcess,
    input_names=("data",),
    output_names=("data",),
    optional_input_names=("serialize_output_data",),
):
    """
    Process used to compute the center of rotation of a scan

    :param axis_params: parameters to configure the axis process
    :type: AxisRP
    """

    _CALCULATIONS_METHODS = {
        AxisMode.centered: compute_scan_cor_nabu_centered,
        AxisMode.global_: compute_scan_cor_nabu_global,
        AxisMode.sliding_window_sinogram: compute_scan_cor_nabu_sliding_window,
        AxisMode.sliding_window_radios: compute_scan_cor_nabu_sliding_window,
        AxisMode.growing_window_sinogram: compute_scan_cor_nabu_growing_window,
        AxisMode.growing_window_radios: compute_scan_cor_nabu_growing_window,
        AxisMode.sino_coarse_to_fine: compute_scan_sino_coarse_to_fine,
        AxisMode.composite_coarse_to_fine: compute_scan_composite_coarse_to_fine,
        AxisMode.near: compute_scan_composite_coarse_to_fine,
        AxisMode.sino_fourier_angles: compute_scan_fourier_angles,
        AxisMode.octave_accurate_radios: compute_scan_octave_accurate,
    }

    def __init__(
        self,
        process_id=None,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        Task.__init__(
            self,
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
        )
        if "recons_params" in inputs:
            raise KeyError("do not use 'recons_params' but use 'axis_params' instead")
        axis_params = inputs.get("axis_params", None)
        if isinstance(axis_params, dict):
            axis_params = AxisRP.from_dict(axis_params)
        elif not (axis_params is None or isinstance(axis_params, AxisRP)):
            raise TypeError(
                "'axis_params' is expected to be None or an instance of AxisRP or a dict"
            )

        SuperviseProcess.__init__(self, process_id=process_id)
        self._mode_calculation_fct = {}
        """dict with function pointer to call for making the mode calculation.
        Function should have only one 'scan' parameter as input"""

        self._axis_params = axis_params or AxisRP()
        """Axis reconstruction parameters to apply"""
        self._locked = False
        """Boolean used to lock reconstruction parameters edition"""
        self._recons_params_before_lock = None
        """Recons parameters register before locking the position"""

    def set_configuration(self, configuration):
        if "_rpSetting" in configuration:
            recons_params = AxisRP.from_dict(configuration["_rpSetting"])
        else:
            recons_params = AxisRP.from_dict(configuration)
        self.set_recons_params(recons_params=recons_params)

    def set_recons_params(self, recons_params):
        assert isinstance(recons_params, AxisRP)
        self._axis_params = recons_params

    def lock_position_value(self, lock=True):
        """
        lock the position currently computed or defined by the user.
        In this case we will lock the axis as defined 'fixed' with the current
        value

        :param bool lock: if true lock the currently existing position value
        """
        self._locked = lock
        if lock:
            self._recons_params_before_lock = self._axis_params.to_dict()
            if self._axis_params not in (AxisMode.manual, AxisMode.read):
                self._axis_params.mode = AxisMode.manual
        else:
            if self._recons_params_before_lock:
                self._axis_params.load_from_dict(
                    self._recons_params_before_lock
                )  # noqa

    def run(self):
        """
        Compute the position value then get ready to the next. And call

        .. note:: this simply call `compute`.
                  But this is needed for the AxisProcessThreaded class
        """
        scan = data_identifier_to_scan(self.inputs.data)
        if scan is None:
            self.outputs.data = None
            return

        if isinstance(scan, TomwerScanBase):
            scan = scan
        elif isinstance(scan, dict):
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            raise ValueError(f"input type {scan} is not managed")

        _logger.info("start axis calculation for %s" % scan.path)
        self._axis_params.frame_width = scan.dim_1
        cor = error = None
        try:
            scan_res = self.compute(scan=scan)
        except Exception as e:
            scan_res = None
            error = e
        else:
            if isinstance(scan_res, TomwerScanBase):
                cor = scan_res.axis_params.relative_cor_value
            elif scan_res is None:
                if scan.axis_params.relative_cor_value is not None:
                    cor = scan.axis_params.relative_cor_value
            elif isinstance(scan_res, float):
                cor = scan_res
            else:
                assert isinstance(scan_res, dict)
                b_dict = scan_res
                if TomwerScanBase._DICT_AXIS_KEYS in scan_res:
                    b_dict = scan_res["axis_params"]
                cor = b_dict["POSITION_VALUE"]
        finally:
            if cor != "...":
                self._process_end(scan, cor=cor, error=error)

        if self.get_input_value("serialize_output_data", True):
            self.outputs.data = scan_res.to_dict()
        else:
            self.outputs.data = scan_res

    def _process_end(self, scan, cor, error=None):
        assert isinstance(scan, TomwerScanBase)
        if scan.process_file is not None:
            entry = "entry"
            if isinstance(scan, NXtomoScan):
                entry = scan.entry
            self.register_process(
                process_file=scan.process_file,
                entry=entry,
                results={"center_of_rotation": cor if cor is not None else "-"},
                configuration=self._axis_params.to_dict(),
                process_index=scan.pop_process_index(),
                overwrite=True,
            )

        try:
            extra = {
                logconfig.DOC_TITLE: self._scheme_title,
                logconfig.SCAN_ID: scan.path,
            }
            if error is not None:
                info = " ".join(
                    (
                        "fail to compute axis position for scan",
                        str(scan.path),
                        "reason is ",
                        str(error),
                    )
                )
                _logger.processFailed(info, extra=extra)
                ProcessManager().notify_dataset_state(
                    dataset=scan, process=self, state=DatasetState.FAILED, details=info
                )
            elif scan.axis_params.relative_cor_value is None:
                info = " ".join(
                    ("fail to compute axis position for scan", str(scan.path))
                )
                _logger.processFailed(info, extra=extra)
                ProcessManager().notify_dataset_state(
                    dataset=scan, process=self, state=DatasetState.FAILED, details=info
                )
            else:
                info = "axis calculation defined for {}: {} (using {})".format(
                    str(scan.path),
                    str(scan.axis_params.relative_cor_value),
                    scan.axis_params.mode.value,
                )
                _logger.processSucceed(info, extra=extra)
                ProcessManager().notify_dataset_state(
                    dataset=scan, process=self, state=DatasetState.SUCCEED, details=info
                )
        except Exception as e:
            _logger.error(e)

    @staticmethod
    def get_inputs_urls(scan):
        """Make sure we have valid projections to be used for axis calculation

        :param TomwerScanBase scan: scan to check
        :raise: NoAxisUrl if fails to found
        :return: the two projections to be used for axis calculation
        :rtype: tuple of AxisResource
        """
        if (
            scan.axis_params
            and scan.axis_params.axis_url_1
            and scan.axis_params.axis_url_1.url
        ):
            return scan.axis_params.axis_url_1, scan.axis_params.axis_url_2
        else:
            _radio_1, _radio_2 = scan.get_opposite_projections(
                mode=scan.axis_params.angle_mode
            )
        return _radio_1, _radio_2

    @staticmethod
    def get_inputs(scan):
        assert isinstance(scan, TomwerScanBase)
        radio_1, radio_2 = AxisTask.get_inputs_urls(scan=scan)
        if radio_1 and radio_2:
            mess = " ".join(
                ("input radios are", radio_1.url.path(), "and", radio_2.url.path())
            )
            _logger.info(mess)
            log_ = scan.axis_params.projection_type is ProjectionType.transmission

            # if necessary normalize data
            if radio_1.normalized_data is None:
                radio_1.normalize_data(scan, log_=log_)
            if radio_2.normalized_data is None:
                radio_2.normalize_data(scan, log_=log_)

            if scan.axis_params.paganin_preproc:
                data_1 = radio_1.normalized_data_paganin
                data_2 = radio_2.normalized_data_paganin
            else:
                data_1 = radio_1.normalized_data
                data_2 = radio_2.normalized_data

            if scan.axis_params.scale_img2_to_img1:
                data_2 = image.scale_img2_to_img1(img_1=data_1, img_2=data_2)
            return data_1, data_2
        else:
            _logger.info("fail to find any inputs")
            return None, None

    def compute(self, scan, wait=True):
        """
        Compute the position value for the scan

        :param TomwerScanBase scan:
        :param bool wait: used for threaded process. True if we want to end the
                          computation before releasing hand.
        :return: scan as a TomoBase
        """
        assert scan is not None
        if isinstance(scan, dict):
            _logger.warning("convert scan from a dict")
            _scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            _scan = scan
        assert isinstance(_scan, TomwerScanBase)
        # if the scan has no tomo reconstruction parameters yet create them
        if _scan.axis_params is None:
            _scan.axis_params = AxisRP()

        # copy axis recons parameters. We skip the axis_url which are specific
        # to the scan
        _scan.axis_params.copy(
            self._axis_params, copy_axis_url=False, copy_flip_lr=False
        )
        assert scan.axis_params is not None
        return self._process_computation(scan=_scan)

    def scan_ready(self, scan):
        _logger.info(scan, "processed")

    def _process_computation(self, scan):
        """

        :param TomwerScanBase scan: scan for which we want to compute the axis
                              position.
        :return: scan as a TomoBase or a dict if serialize_output_data activated
        """
        _logger.info("compute center of rotation for %s" % scan.path)
        try:
            position = self.compute_axis_position(scan)
        except NotImplementedError as e:
            scan.axis_params.set_relative_value(None)
            raise e
        except ValueError as e:
            scan_name = scan.path or "undef scan"
            scan.axis_params.set_relative_value(None)
            raise Exception(
                f"Fail to compute axis position for {scan_name} reason is {e}"
            )
        else:
            scan.axis_params.set_relative_value(position)
            self._axis_params.frame_width = scan.dim_1
            self._axis_params.set_relative_value(position)
            scan_name = scan.path or "undef scan"
            if scan.axis_params.use_sinogram:
                method = "sinogram"
            else:
                method = scan.axis_params.mode.value
            r_cor_value = scan.axis_params.relative_cor_value
            mess = (
                f"Compute axis position ({r_cor_value}) with {method} for {scan_name}"
            )
            _logger.info(mess)
        return scan

    def setMode(self, mode, value):
        if mode is AxisMode.manual:
            self._axis_params.cor_position = value
        else:
            raise NotImplementedError("mode not implemented yet")

    def define_calculation_mode(self, mode, fct_pointer):
        """Register the function to call of the given mode

        :param AxisMode mode: the mode to register
        :param fct_pointer: pointer to the function to call
        """
        self._mode_calculation_fct[mode] = fct_pointer

    def compute_axis_position(self, scan):
        """

        :param scan: scan for which we compute the center of rotation
        :type: TomoScan
        :return: position of the rotation axis. Use the `.AxisMode` defined
                 by the `.ReconsParams` of the `.AxisProcess`
        :rtype: float or None (if fail to compute the axis position)
        """
        mode = self._axis_params.mode
        if mode in (AxisMode.manual, AxisMode.read):
            # If mode is read or manual the position_value is not computed and
            # we will keep the actual one (should have been defined previously)
            res = self._axis_params.relative_cor_value
        elif mode in self._CALCULATIONS_METHODS:
            _logger.info("use radios, mode is %s" % mode.value)
            res = self._CALCULATIONS_METHODS[mode](scan)
        else:
            raise NotImplementedError("Method for", mode, "is not defined")
        return res

    @docstring(Task.program_name)
    @staticmethod
    def program_name():
        return "tomwer_axis"

    @docstring(Task.program_version)
    @staticmethod
    def program_version():
        return tomwer.version.version

    @docstring(Task.definition)
    @staticmethod
    def definition():
        return "Compute center of rotation"

    @staticmethod
    def get_cor_frm_process_file(
        process_file, entry, as_url=False
    ) -> Union[None, float]:
        """
        Read cor position from a tomwer_process file

        :param process_file:
        :param entry:
        :return:
        """
        if entry is None:
            with open_hdf5(process_file) as h5f:
                entries = AxisTask._get_process_nodes(root_node=h5f, process=AxisTask)
                if len(entries) == 0:
                    _logger.info("unable to find a Axis process in %s" % process_file)
                    return None
                elif len(entries) > 1:
                    raise ValueError("several entry found, entry should be " "specify")
                else:
                    entry = list(entries.keys())[0]
                    _logger.info("take %s as default entry" % entry)

        with open_hdf5(process_file) as h5f:
            axis_nodes = AxisTask._get_process_nodes(
                root_node=h5f[entry], process=AxisTask
            )
            index_to_path = {}
            for key, index in axis_nodes.items():
                index_to_path[index] = key

            if len(axis_nodes) == 0:
                return None
            # take the last processed dark ref
            last_process_index = sorted(list(axis_nodes.values()))[-1]
            last_process_dark = index_to_path[last_process_index]
            if (len(index_to_path)) > 1:
                _logger.debug(
                    "several processing found for dark-ref,"
                    "take the last one: %s" % last_process_dark
                )

            res = None
            if "results" in h5f[last_process_dark].keys():
                results_node = h5f[last_process_dark]["results"]
                if "center_of_rotation" in results_node.keys():
                    if as_url:
                        res = DataUrl(
                            file_path=process_file,
                            data_path="/".join((results_node, "center_of_rotation")),
                            scheme="h5py",
                        )
                    else:
                        res = h5py_read_dataset(results_node["center_of_rotation"])
            return res


class AxisProcess(AxisTask):
    def __init__(
        self,
        process_id=None,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        deprecated_warning(
            name="tomwer.core.process.reconstruction.axis.axis.AxisProcess",
            type_="class",
            reason="improve readibility",
            since_version="1.2",
            replacement="AxisTask",
        )
        super().__init__(process_id, varinfo, inputs, node_id, node_attrs, execinfo)



# coding: utf-8
###########################################################################
# Copyright (C) 2016-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "19/07/2021"


import functools
import logging

from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess
from silx.gui import qt

from tomwer.core.process.reconstruction.normalization import SinoNormalizationTask
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.settings import get_lbsram_path, isOnLbsram
from tomwer.core.utils.lbsram import is_low_on_memory

from ..processingstack import FIFO, ProcessingThread

_logger = logging.getLogger(__name__)


class INormalizationProcessStack(FIFO, qt.QObject):
    """Implementation of the `.AxisProcess` but having a stack for treating
    scans and making computation in threads"""

    def __init__(self, process_id=None):
        qt.QObject.__init__(self)
        FIFO.__init__(self, process_id=process_id)

    def _process(self, data, configuration, callback=None):
        ProcessManager().notify_dataset_state(
            dataset=data,
            process=self,
            state=DatasetState.ON_GOING,
        )
        _logger.processStarted(f"start intensity normalization {data}")
        assert isinstance(data, TomwerScanBase)
        if isOnLbsram(data) and is_low_on_memory(get_lbsram_path()) is True:
            # if computer is running into low memory on lbsram skip it
            mess = "low memory, skip intensity normalization", data.path
            ProcessManager().notify_dataset_state(
                dataset=data, process=self._process_id, state=DatasetState.SKIPPED
            )
            _logger.processSkipped(mess)
            if callback is not None:
                callback()
            self.scan_ready(scan=data)
        else:
            self._data_currently_computed = data
            self._computationThread.init(data=data, i_norm_params=configuration)
            # need to manage connect before starting it because
            fct_callback = functools.partial(self._end_threaded_computation, callback)
            self._computationThread.finished.connect(fct_callback)
            self._computationThread.start()

    def _end_threaded_computation(self, callback=None):
        assert self._data_currently_computed is not None
        self._computationThread.finished.disconnect()
        if callback:
            callback()
        FIFO._end_threaded_computation(self)

    def _create_processing_thread(self, process_id=None) -> qt.QThread:
        return _ProcessingThread(process_id=process_id)


class _ProcessingThread(ProcessingThread, SuperviseProcess):
    """
    Thread use to execute the processing of intensity normalization
    (mostly load io and compute the mean / median on it)
    """

    def __init__(self, process_id=None):
        SuperviseProcess.__init__(self, process_id=process_id)
        ProcessingThread.__init__(self, process_id=process_id)
        self.center_of_rotation = None
        self._dry_run = False
        self._scan = None
        self._i_norm_params = None

    def init(self, data, i_norm_params):
        self._scan = data
        self._i_norm_params = i_norm_params

    def run(self):
        self.sigComputationStarted.emit()
        norm_process = SinoNormalizationTask(
            process_id=self.process_id,
            varinfo=None,
            inputs={
                "data": self._scan,
                "configuration": self._i_norm_params,
                "serialize_output_data": False,
            },
        )
        try:
            norm_process.run()
        except Exception as e:
            _logger.error(str(e))
            mess = f"Intensity normalization computation for {self._scan} failed."
            state = DatasetState.FAILED
        else:
            mess = f"Intensity normalization computation for {self._scan} succeed."
            state = DatasetState.WAIT_USER_VALIDATION

        ProcessManager().notify_dataset_state(
            dataset=self._scan,
            process=self,
            state=state,
            details=mess,
        )



from __future__ import annotations
import os
import logging

from tomwer.core.process.task import Task
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.volume.hdf5volume import HDF5VolumeIdentifier
from tomwer.core.volume.rawvolume import RawVolumeIdentifier
from tomwer.io.utils.raw_and_processed_data import file_is_on_processed_data
from tomoscan.identifier import VolumeIdentifier
from tomoscan.esrf.volume.singleframebase import VolumeSingleFrameBase

from processview.core.manager import DatasetState, ProcessManager

try:
    from pyicat_plus.client.main import IcatClient  # noqa F401
except ImportError:
    has_pyicat_plus = False
else:
    has_pyicat_plus = True

_logger = logging.getLogger(__name__)


class PublishProcessedDataFolderTask(
    Task,
    input_names=(
        "beamline",
        "proposal",
        "dataset",
        "path",
        "raw",
    ),
    optional_input_names=(
        "metadata",
        "dry_run",
    ),
):
    """publish a list of volume to icat for the provided beamline && dataset && proposal"""

    def run(self):
        beamline = self.inputs.beamline
        proposal = self.inputs.proposal
        dataset = self.inputs.dataset
        raw = self.inputs.raw
        metadata = self.get_input_value("metadata", {})
        path = self.inputs.path
        dry_run = self.get_input_value("dry_run", False)

        # checks (need because can be pass by a GUI and will set those values to empty string mostly)
        missing = []
        if beamline in (None, ""):
            missing.append("beamline")
        if proposal in (None, ""):
            missing.append("proposal")
        if dataset in (None, ""):
            missing.append("dataset")
        if raw is None or len(raw) == 0:
            missing.append("raw")

        if len(missing) > 0:
            mess = f"Missing information about {','.join(missing)}"
            _logger.error(mess)
            raise ValueError(mess)
        if not os.path.exists(path):
            raise ValueError(f"path: {path} doesn't exists")

        if not file_is_on_processed_data(str(path)):
            # for now it is safer to limit the publication of metadata to the folder contained in 'PROCESSED_DATA'
            raise ValueError(
                f"processed data dir ({path}) is not in a PROCESSED_DATA folder. Cannot publish it"
            )
        # publish
        if not has_pyicat_plus:
            raise ImportError("pyicat_plus not installed")

        if not dry_run:
            print("will publish")
            from pprint import pprint

            pprint(
                {
                    "beamline": beamline,
                    "proposal": proposal,
                    "dataset": dataset,
                    "path": path,
                    "metadata": metadata,
                    "raw": raw,
                }
            )
            icat_client = IcatClient(
                metadata_urls=("bcu-mq-01.esrf.fr:61613", "bcu-mq-02.esrf.fr:61613")
            )

            icat_client.store_processed_data(
                beamline=beamline,
                proposal=proposal,
                dataset=dataset,
                path=path,
                metadata=metadata,
                raw=raw,
            )


class PublishReconstructedVolumeFromScanTask(
    Task,
    input_names=(
        "data",
        "beamline",
        "proposal",
        "dataset",
    ),
    optional_input_names=(
        "metadata",
        "dry_run",
        "__process__",
    ),
):
    """
    Proxy to PublishProcessedDataTask
    publish processed data from a 'TomwerScanBase' instance. We expect volume to be registered under 'latest_vol_reconstructions' property.
    """

    def __init__(
        self, varinfo=None, inputs=None, node_id=None, node_attrs=None, execinfo=None
    ):
        super().__init__(varinfo, inputs, node_id, node_attrs, execinfo)
        scan = self.inputs.data

        if not isinstance(scan, TomwerScanBase):
            raise TypeError(f"scan is expected to be an instance of {TomwerScanBase}")
        if isinstance(scan, EDFTomoScan):
            raise TypeError(
                "EDFTomoScan is not handling for pushing processed data to icat"
            )
        else:
            raw = from_bliss_original_file_to_raw(scan.get_bliss_orginal_files())

        def get_volume_identifier_path_to_save(volume_identfier):
            assert isinstance(volume_identfier, VolumeIdentifier)
            if isinstance(
                volume_identfier, (HDF5VolumeIdentifier, RawVolumeIdentifier)
            ):
                return os.path.dirname(volume_identfier.file_path)
            elif isinstance(volume_identfier, VolumeSingleFrameBase):
                if volume_identfier.url is not None:
                    return volume_identfier.url.file_path()
                elif volume_identfier.data_url.file_path() is not None:
                    return volume_identfier.data_url.file_path()
            else:
                raise ValueError(
                    f"identifier type {type(volume_identfier)} is not handled"
                )

        paths_to_save = set(
            [
                get_volume_identifier_path_to_save(volume_id)
                for volume_id in scan.latest_vol_reconstructions
            ]
        )
        self._sub_tasks = [
            PublishProcessedDataFolderTask(
                inputs={
                    "beamline": self.inputs.beamline,
                    "proposal": self.inputs.proposal,
                    "dataset": self.inputs.dataset,
                    "path": path_to_save,
                    "raw": raw,
                    "metadata": scan.build_icat_metadata(),
                    "dry_run": self.get_input_value("dry_run", False),
                }
            )
            for path_to_save in paths_to_save
        ]

    def run(self):
        scan = self.inputs.data
        process = self.get_input_value("__process__", None)

        if process is not None:
            ProcessManager().notify_dataset_state(
                dataset=scan,
                process=process(),
                state=DatasetState.ON_GOING,
            )

        try:
            for sub_task in self._sub_tasks:
                sub_task.run()
        except Exception as e:
            if process is not None:
                ProcessManager().notify_dataset_state(
                    dataset=scan,
                    process=process(),
                    state=DatasetState.FAILED,
                )
                raise e
        else:
            if process is not None:
                ProcessManager().notify_dataset_state(
                    dataset=scan,
                    process=process(),
                    state=DatasetState.SUCCEED,
                )


def from_bliss_original_file_to_raw(bliss_original_files: tuple | None) -> tuple:
    """
    convert NXtomo 'bliss_original_files' to icat raw parameter (folder containing the raw)
    without '/mnt/multipath-shares' prefix
    """
    if bliss_original_files is None:
        return None

    # TODO: FIXME
    def fix_path(file_path):
        path = os.path.dirname(file_path)
        if "/mnt/multipath-shares" in path:
            # no simple workaround. abspath return a path with '/mnt/multipath-shares'
            _logger.info(
                "looks like raw data is given with '/mnt/multipath-shares' prefix. Icat will fail on it. Must remove it. No proper other handling found :()"
            )
            # small workaround to fix abspath. Should not be the case anymore so raise an error
            path.replace("/mnt/multipath-shares", "")
        return path

    return tuple([os.path.dirname(bliss_file) for bliss_file in bliss_original_files])



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "25/08/2022"

import logging

import tomoscan.esrf.scan.utils
from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output, OWBaseWidget

import tomwer.core.process.control.singletomoobj
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.tomwer_object import TomwerObject
from tomwer.core.volume.volumefactory import VolumeFactory
from tomwer.gui.control.singletomoobj import SingleTomoObj

_logger = logging.getLogger(__name__)


class SingleTomoObjOW(OWBaseWidget, openclass=True):
    name = "single tomo obj"
    id = "orange.widgets.tomwer.control.SingleScanOW.SingleScanOW"
    description = "Definition of a single dataset"
    icon = "icons/single_tomo_obj.svg"
    priority = 51
    keywords = ["tomography", "NXtomo", "tomwer", "folder", "scan"]

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    ewokstaskclass = tomwer.core.process.control.singletomoobj.SingleTomoObjProcess

    _tomo_obj_setting = Setting(str())

    class Inputs:
        tomo_obj = Input(name="tomo_obj", type=TomwerObject)

    class Outputs:
        tomo_obj = Output(
            name="tomo_obj", type=TomwerObject, doc="one object to be process"
        )
        reduced_darks = Output(
            name="reduced dark(s)",
            type=dict,
            doc="Reduced darks as a dict",
        )
        reduced_flats = Output(
            name="reduced flat(s)",
            type=dict,
            doc="Reduced flats as a dict",
        )
        relative_reduced_darks = Output(
            name="relative reduced dark(s)",
            type=dict,
            doc="Reduced darks as a dict. Indexes are provided as relative so in [0.0, 1.0[",
        )
        relative_reduced_flats = Output(
            name="relative reduced flat(s)",
            type=dict,
            doc="Reduced flats as a dict. Indexes are provided as relative so in [0.0, 1.0[",
        )

    def __init__(self, parent=None):
        super().__init__(parent)
        self._latest_received_scan = None
        # small work around to keep in scan processing cache and avoid recomputing it if not necessary
        self.widget = SingleTomoObj(parent=self)
        self.widget.sigTomoObjChanged.connect(self._updateSettings)
        self.widget.sigTomoObjChanged.connect(self._triggerObjDownstream)

        self._loadSettings()
        layout = gui.vBox(self.mainArea, self.name).layout()
        layout.addWidget(self.widget)

    @Inputs.tomo_obj
    def add(self, tomo_obj):
        if tomo_obj is None:
            return
        self.widget.setTomoObject(tomo_obj)

    def _loadSettings(self):
        if self._tomo_obj_setting != str():
            self.widget.setTomoObject(self._tomo_obj_setting)

    def _updateSettings(self):
        self._tomo_obj_setting = self.widget.getTomoObjIdentifier()

    def _triggerObjDownstream(self):
        obj_identifier = self.widget.getTomoObjIdentifier()
        if (
            self._latest_received_scan is not None
            and self._latest_received_scan.get_identifier() == obj_identifier
        ):
            data = self._latest_received_scan
        else:
            try:
                data = VolumeFactory.create_tomo_object_from_identifier(obj_identifier)
            except:  # noqa E722
                try:
                    data = ScanFactory.create_tomo_object_from_identifier(
                        obj_identifier
                    )
                except:  # noqa E722
                    _logger.warning(
                        f"Unable to find an obj with {obj_identifier} as identifier"
                    )
                    return

        self.Outputs.tomo_obj.send(data)
        # if the data is a scan then provide also access to the reduced frames. Can be convenient
        if isinstance(data, TomwerScanBase):
            if data.reduced_darks not in (None, {}):
                reduced_darks = data.reduced_darks
                self.Outputs.reduced_darks.send(reduced_darks)

                # we want to send those in relative position to have something generic. This is a convention for now
                reduced_darks = (
                    tomoscan.esrf.scan.utils.from_absolute_reduced_frames_to_relative(
                        reduced_frames=data.reduced_darks, scan=data
                    )
                )
                reduced_darks["reduce_frames_name"] = (
                    f"darks from {data.get_identifier().short_description()}"
                )
                self.Outputs.relative_reduced_darks.send(reduced_darks)

            if data.reduced_flats not in (None, {}):
                reduced_flats = data.reduced_flats
                self.Outputs.reduced_flats.send(reduced_flats)

                # we want to send those in relative position to have something generic. This is a convention for now
                reduced_flats = (
                    tomoscan.esrf.scan.utils.from_absolute_reduced_frames_to_relative(
                        reduced_frames=data.reduced_flats, scan=data
                    )
                )
                reduced_flats["reduce_frames_name"] = (
                    f"flats from {data.get_identifier().short_description()}"
                )
                self.Outputs.relative_reduced_flats.send(reduced_flats)



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/


__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "13/07/2022"


import logging
from typing import Optional, Union
from urllib.parse import urlparse

from tomoscan import identifier as _identifier_mod
from tomoscan.esrf.identifier.url_utils import split_path
from tomoscan.factory import Factory as _oVolumeFactory
from tomoscan.identifier import BaseIdentifier, ScanIdentifier, VolumeIdentifier

from tomwer.core.scan.edfscan import EDFTomoScan, EDFTomoScanIdentifier
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.volume import (
    EDFVolume,
    HDF5Volume,
    JP2KVolume,
    MultiTIFFVolume,
    RawVolume,
    TIFFVolume,
)
from tomwer.core.volume.edfvolume import EDFVolumeIdentifier
from tomwer.core.volume.hdf5volume import HDF5VolumeIdentifier
from tomwer.core.volume.jp2kvolume import JP2KVolumeIdentifier
from tomwer.core.volume.rawvolume import RawVolumeIdentifier
from tomwer.core.volume.tiffvolume import (
    MultiTiffVolumeIdentifier,
    TIFFVolumeIdentifier,
)
from tomwer.utils import docstring

_logger = logging.getLogger(__name__)


class VolumeFactory(_oVolumeFactory):
    @staticmethod
    def from_identifier_to_vol_urls(identifier) -> Optional[tuple]:
        """
        convert an identifier to a volume
        and return all the existing url of this volume
        """
        try:
            vol = VolumeFactory.create_tomo_object_from_identifier(
                identifier=identifier
            )
        except Exception as e:
            _logger.error(e)
            return None
        else:
            return tuple(vol.browse_data_urls())

    @docstring(_oVolumeFactory.create_tomo_object_from_identifier)
    @staticmethod
    def create_tomo_object_from_identifier(identifier: Union[str, BaseIdentifier]):
        """

        :param TextIOWrapper scan_path: path to the scan directory or file
        :param entry: entry on the file. Requested for hdf5 files
        :param accept_bliss_scan: if True the factory can return some BlissScan
                                  But this is only compatible with the
                                  Tomomill processing.
        :return: TomwerScanBase instance fitting the scan folder or scan path
        :rtype: tomwer.core.scan.scanbase.TomwerScanBase
        :raises: ValueError if scan_path is not containing a scan
        """
        if isinstance(identifier, BaseIdentifier):
            identifier = identifier.to_str()
        elif not isinstance(identifier, str):
            raise TypeError(
                f"identifier is expected to be a str or an instance of {BaseIdentifier} not {type(identifier)}"
            )

        # step 1: convert identifier to an instance of BaseIdentifier if necessary
        if isinstance(identifier, str):
            info = urlparse(identifier)
            paths = split_path(info.path)
            scheme = info.scheme
            if len(paths) == 1:
                # insure backward compatibility. Originally (until 0.8) there was only one type which was scan
                tomo_type = ScanIdentifier.TOMO_TYPE
            elif len(paths) == 2:
                tomo_type, _ = paths
            else:
                raise ValueError("Failed to parse path string:", info.path)

            if tomo_type == _identifier_mod.VolumeIdentifier.TOMO_TYPE:
                if scheme == "edf":
                    identifier = EDFVolumeIdentifier.from_str(identifier=identifier)
                elif scheme == "hdf5":
                    identifier = HDF5VolumeIdentifier.from_str(identifier=identifier)
                elif scheme == "tiff":
                    identifier = TIFFVolumeIdentifier.from_str(identifier=identifier)
                elif scheme == "tiff3d":
                    identifier = MultiTiffVolumeIdentifier.from_str(
                        identifier=identifier
                    )
                elif scheme == "jp2k":
                    identifier = JP2KVolumeIdentifier.from_str(identifier=identifier)
                elif scheme in ("raw", "vol"):
                    identifier = RawVolumeIdentifier.from_str(identifier=identifier)
                else:
                    raise ValueError(f"Scheme {scheme} is not recognized")

            elif tomo_type == _identifier_mod.ScanIdentifier.TOMO_TYPE:
                # otherwise consider this is a scan. Insure backward compatibility
                if scheme == "edf":
                    identifier = EDFTomoScanIdentifier.from_str(identifier=identifier)
                elif scheme == "hdf5":
                    identifier = HDF5VolumeIdentifier.from_str(identifier=identifier)
                else:
                    raise ValueError(f"Scheme {scheme} not recognized")
            else:
                raise ValueError(f"{tomo_type} is not an handled tomo type")

        # step 2: convert identifier to a TomoBaseObject
        assert isinstance(identifier, BaseIdentifier)
        scheme = identifier.scheme
        tomo_type = identifier.tomo_type

        if scheme == "edf":
            if tomo_type == VolumeIdentifier.TOMO_TYPE:
                return EDFVolume.from_identifier(identifier=identifier)
            elif tomo_type == ScanIdentifier.TOMO_TYPE:
                return EDFTomoScan.from_identifier(identifier=identifier)
            else:
                raise NotImplementedError()
        elif scheme == "hdf5":
            if tomo_type == VolumeIdentifier.TOMO_TYPE:
                return HDF5Volume.from_identifier(identifier=identifier)
            elif tomo_type == ScanIdentifier.TOMO_TYPE:
                return NXtomoScan.from_identifier(identifier=identifier)
            else:
                raise NotImplementedError()
        elif scheme == "jp2k":
            if tomo_type == VolumeIdentifier.TOMO_TYPE:
                return JP2KVolume.from_identifier(identifier=identifier)
            else:
                raise NotImplementedError
        elif scheme == "tiff":
            if tomo_type == VolumeIdentifier.TOMO_TYPE:
                return TIFFVolume.from_identifier(identifier=identifier)
            else:
                raise NotImplementedError
        elif scheme == "tiff3d":
            if tomo_type == VolumeIdentifier.TOMO_TYPE:
                return MultiTIFFVolume.from_identifier(identifier=identifier)
            else:
                raise NotImplementedError
        elif scheme in ("raw", "vol"):
            if tomo_type == VolumeIdentifier.TOMO_TYPE:
                return RawVolume.from_identifier(identifier=identifier)
            else:
                raise NotImplementedError
        else:
            raise ValueError(f"Scheme {scheme} not recognized")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = [
    "H. Payno",
]
__license__ = "MIT"
__date__ = "12/07/2022"

import logging

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output, OWBaseWidget
from silx.gui import qt

import tomwer.core.process.control.volumeselector
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.volume.volumebase import TomwerVolumeBase
from tomwer.gui.control.volumeselectorwidget import VolumeSelectorWidget

logger = logging.getLogger(__name__)


class VolumeSelectorOW(OWBaseWidget, openclass=True):
    name = "volume selector"
    id = "orange.widgets.tomwer.volumeselector"
    description = (
        "List all received volumes. Then user can select a specific"
        "volume to be passed to the next widget."
    )
    icon = "icons/volumeselector.svg"
    priority = 62
    keywords = ["tomography", "selection", "tomwer", "volume"]

    ewokstaskclass = (
        tomwer.core.process.control.volumeselector._VolumeSelectorPlaceHolder
    )

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    _scanIDs = Setting(list())

    class Inputs:
        volume = Input(name="volume", type=TomwerVolumeBase, multiple=True)

    class Outputs:
        volume = Output(name="volume", type=TomwerVolumeBase)

    def __init__(self, parent=None):
        """ """
        super().__init__(parent)

        self.widget = VolumeSelectorWidget(parent=self)
        self._loadSettings()

        self.widget.sigUpdated.connect(self._updateSettings)
        self.widget.sigSelectionChanged.connect(self.changeSelection)
        layout = gui.vBox(self.mainArea, self.name).layout()
        layout.addWidget(self.widget)
        # expose API
        self.setActiveScan = self.widget.setActiveData
        self.selectAll = self.widget.selectAll
        self.add = self.widget.add

    @Inputs.volume
    def _volumeReceived(self, volume, *args, **kwargs):
        self.addVolume(volume)

    def addVolume(self, volume):
        if volume is not None:
            self.widget.add(volume)

    def removeVolume(self, volume):
        if volume is not None:
            self.widget.remove(volume)

    def changeSelection(self, list_volume):
        if list_volume:
            for volume_id in list_volume:
                volume = self.widget.dataList.getVolume(volume_id, None)
                if volume is not None:
                    assert isinstance(volume, TomwerVolumeBase)
                    self.Outputs.volume.send(volume)
                else:
                    logger.error(f"{volume_id} not found the list")

    def send(self):
        """send output signals for each selected items"""
        sItem = self.widget.dataList.selectedItems()
        if sItem and len(sItem) >= 1:
            selection = [_item.text() for _item in sItem]
            self.changeSelection(list_volume=selection)

    def _loadSettings(self):
        for scan in self._scanIDs:
            assert isinstance(scan, str)
            # kept for backward compatibility since 0.11. To be removed on the future version.
            if "@" in scan:
                entry, file_path = scan.split("@")
                nxtomo_scan = NXtomoScan(entry=entry, scan=file_path)
                self.addVolume(nxtomo_scan)
            else:
                self.addVolume(scan)

    def _updateSettings(self):
        self._scanIDs = []
        for scan in self.widget.dataList._myitems:
            self._scanIDs.append(scan)

    def keyPressEvent(self, event):
        """
        To shortcut orange and make sure the `delete` key will be interpreted we need to overwrite this function
        """
        if event.key() == qt.Qt.Key_Delete:
            self.widget._callbackRemoveSelectedDatasets()
        else:
            super().keyPressEvent(event)



from __future__ import annotations
from silx.gui import qt
from nabu.stitching.sample_normalization import (
    SampleSide as _SampleSide,
    Method as _SampleNormalizationMethod,
)
from nabu.stitching import config as _config_stitching
from tomwer.gui.utils.qt_utils import block_signals


class NormalizationBySampleGroupBox(qt.QGroupBox):
    """
    Widget to define the normalization to apply to frames"""

    sigConfigChanged = qt.Signal()
    """Emit when the configuration is changed"""

    def __init__(self, title: str = "normalization by sample", parent=None):
        # FIXME: add a way to the user for requesting a view of the region picked ???
        super().__init__(title, parent)
        self.setCheckable(True)
        self.setLayout(qt.QFormLayout())

        # method
        self._methodCB = qt.QComboBox(self)
        self._methodCB.addItems(_SampleNormalizationMethod.values())
        self._methodCB.setCurrentText(_SampleNormalizationMethod.MEDIAN.value)
        self.layout().addRow("method", self._methodCB)

        # side
        self._sideCB = qt.QComboBox(self)
        self._sideCB.addItems(_SampleSide.values())
        self._sideCB.setCurrentText(_SampleSide.LEFT.value)
        self.layout().addRow("sampling side", self._sideCB)

        # width
        self._widthSB = qt.QSpinBox(self)
        self._widthSB.setRange(1, 9999999)
        self._widthSB.setValue(30)
        self._widthSB.setSingleStep(10)
        self.layout().addRow("sampling width", self._widthSB)

        # margin
        self._marginSB = qt.QSpinBox(self)
        self._marginSB.setRange(0, 999999)
        self._marginSB.setValue(0)
        self.layout().addRow("sampling margin", self._marginSB)

        # connect signal / slot
        self._methodCB.currentIndexChanged.connect(self._configChanged)
        self._sideCB.currentIndexChanged.connect(self._configChanged)
        self._widthSB.valueChanged.connect(self._configChanged)
        self._marginSB.valueChanged.connect(self._configChanged)

    def _configChanged(self, *args, **kwargs):
        self.sigConfigChanged.emit()

    def getMethod(self) -> _SampleNormalizationMethod:
        return _SampleNormalizationMethod.from_value(self._methodCB.currentText())

    def setMethod(self, method: _SampleNormalizationMethod | str):
        method = _SampleNormalizationMethod.from_value(method)
        self._methodCB.setCurrentText(method.value)

    def getSide(self) -> _SampleSide:
        return _SampleSide.from_value(self._sideCB.currentText())

    def setSide(self, side: _SampleSide):
        side = _SampleSide.from_value(side)
        self._sideCB.setCurrentText(side.value)

    def getMargin(self) -> int:
        return self._marginSB.value()

    def setMargin(self, margin: int):
        self._marginSB.setValue(int(margin))

    def getWidth(self) -> int:
        return self._widthSB.value()

    def setWidth(self, width: int):
        self._widthSB.setValue(int(width))

    def getConfiguration(self) -> dict:
        return {
            _config_stitching.NORMALIZATION_BY_SAMPLE_ACTIVE_FIELD: self.isChecked(),
            _config_stitching.NORMALIZATION_BY_SAMPLE_METHOD: self.getMethod().value,
            _config_stitching.NORMALIZATION_BY_SAMPLE_SIDE: self.getSide().value,
            _config_stitching.NORMALIZATION_BY_SAMPLE_MARGIN: self.getMargin(),
            _config_stitching.NORMALIZATION_BY_SAMPLE_WIDTH: self.getWidth(),
        }

    def setConfiguration(self, config: dict) -> None:
        with block_signals(self):
            method = config.get(_config_stitching.NORMALIZATION_BY_SAMPLE_METHOD, None)
            if method is not None:
                self.setMethod(method=method)

            side = config.get(_config_stitching.NORMALIZATION_BY_SAMPLE_SIDE, None)
            if side is not None:
                self.setSide(side=side)

            margin = config.get(_config_stitching.NORMALIZATION_BY_SAMPLE_MARGIN, None)
            if margin is not None:
                self.setMargin(margin=margin)

            width = config.get(_config_stitching.NORMALIZATION_BY_SAMPLE_WIDTH, None)
            if width is not None:
                self.setWidth(width=width)

            active = config.get(
                _config_stitching.NORMALIZATION_BY_SAMPLE_ACTIVE_FIELD, None
            )
            if active is not None:
                self.setChecked(active in (True, 1, "1", "True"))

        self._configChanged()



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

"""
This module is used to define the process of the reference creator.
This is related to the issue #184
"""

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "21/09/2018"


import logging

from silx.gui import qt
from pyunitsystem import metricsystem

from tomwer.core.utils.char import MU_CHAR

_logger = logging.getLogger(__name__)


class PixelEntry(qt.QWidget):
    valueChanged = qt.Signal()

    class Validator(qt.QIntValidator):
        def validate(self, a0: str, a1: int):
            if a0 == "unknow":
                return qt.QValidator.Acceptable
            else:
                return super().validate(a0, a1)

    def __init__(self, name, parent=None):
        qt.QWidget.__init__(self, parent)

        self.setLayout(qt.QHBoxLayout())
        self._label = qt.QLabel(name, parent=self)
        self.layout().addWidget(self._label)
        self._qlePixelSize = qt.QLineEdit(parent=self)
        self._qlePixelSize.setValidator(self.Validator(self._qlePixelSize))
        self._qlePixelSize.setPlaceholderText("px")
        self.layout().addWidget(self._qlePixelSize)

        # connect signal / slot
        self._qlePixelSize.editingFinished.connect(self.valueChanged)

    def getValue(self):
        if self._qlePixelSize.text() in ("unknown", ""):
            return None
        else:
            return int(self._qlePixelSize.text())

    def setValue(self, value: int):
        self._qlePixelSize.setText(str(value))
        self.valueChanged.emit()

    def setReadOnly(self, read_only: bool) -> None:
        self._qlePixelSize.setReadOnly(read_only)


class MetricEntry(qt.QWidget):
    """
    Create a simple line with a name, a QLineEdit and a combobox to define the
    unit in order to define a metric value.

    :param str name: name of the metric value to define
    :param str: base_unit. Default way to present a value when setted
    """

    editingFinished = qt.Signal()

    class DoubleValidator(qt.QDoubleValidator):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.setNotation(qt.QDoubleValidator.ScientificNotation)

        def validate(self, a0: str, a1: int):
            if a0 == "unknown":
                return (qt.QDoubleValidator.Acceptable, a0, a1)
            else:
                return super().validate(a0, a1)

    _CONVERSION = {
        "nm": metricsystem.nanometer.value,
        f"{MU_CHAR}m": metricsystem.micrometer.value,
        "mm": metricsystem.millimeter.value,
        "cm": metricsystem.centimeter.value,
        "m": metricsystem.meter.value,
    }

    valueChanged = qt.Signal()

    def __init__(self, name, value=0.0, default_unit="m", parent=None):
        qt.QWidget.__init__(self, parent)
        assert type(default_unit) is str
        assert default_unit in ("nm", "mm", "cm", "m", f"{MU_CHAR}m")
        self._base_unit = default_unit

        self.setLayout(qt.QHBoxLayout())
        self._label = qt.QLabel(name, parent=self)
        self.layout().addWidget(self._label)
        self._qlePixelSize = qt.QLineEdit("0.0", parent=self)
        self._qlePixelSize.setValidator(self.DoubleValidator(self._qlePixelSize))
        self.layout().addWidget(self._qlePixelSize)

        self._qcbUnit = qt.QComboBox(parent=self)
        self._qcbUnit.addItem("nm")
        self._qcbUnit.addItem(f"{MU_CHAR}m")
        self._qcbUnit.addItem("mm")
        self._qcbUnit.addItem("cm")
        self._qcbUnit.addItem("m")
        self.layout().addWidget(self._qcbUnit)
        self._resetBaseUnit()

        # connect signal / slot
        self._qcbUnit.currentIndexChanged.connect(self._editingFinished)
        self._qlePixelSize.editingFinished.connect(self._editingFinished)

    def _editingFinished(self, *args, **kwargs):
        self.editingFinished.emit()

    def setReadOnly(self, a0: bool):
        self._qlePixelSize.setReadOnly(a0)
        self._qcbUnit.setEnabled(not a0)

    def setLabelText(self, text: str):
        self._label.setText(text)

    def getCurrentUnit(self):
        assert self._qcbUnit.currentText() in self._CONVERSION
        return self._CONVERSION[self._qcbUnit.currentText()]

    def setValue(self, value_m, displayed_unit: str = "m"):
        """

        :param float value: pixel size in international metric system (meter)
        """
        _value = value_m
        if _value in (None, "unknown"):
            txt = "unknown"
        elif isinstance(_value, str):
            if "..." in _value:
                txt = _value
            else:
                try:
                    _value = float(_value)
                except Exception as error:
                    raise ValueError("Given string does not represent a float", error)
                else:
                    assert isinstance(_value, float)
                    txt = str(_value)
        else:
            txt = str(_value)
        self._qlePixelSize.setText(txt)
        self._resetBaseUnit(displayed_unit=displayed_unit)

    def _resetBaseUnit(self, displayed_unit=None):
        """Simple reset of the combobox according to the base_unit"""
        displayed_unit = displayed_unit or self._base_unit
        index = self._qcbUnit.findText(displayed_unit)
        if index is None:
            raise ValueError("unrecognized base unit")
        else:
            self._qcbUnit.setCurrentIndex(index)

    def getValue(self):
        """

        :return: the value in meter
        :rtype: float
        """
        if self._qlePixelSize.text() in ("unknown", ""):
            return None
        else:
            return float(self._qlePixelSize.text()) * self.getCurrentUnit()

    def setValidator(self, validator):
        self._qlePixelSize.setValidator(validator)

    def setUnit(self, unit):
        unit = str(metricsystem.MetricSystem.from_value(unit))
        idx = self._qcbUnit.findText(unit)
        self._qcbUnit.setCurrentIndex(idx)


class CentimeterEntry(MetricEntry):
    """
    Define a QlineEdit which will set and get metric values in centimeter
    """

    def __init__(self, name, default_unit="cm", parent=None):
        MetricEntry.__init__(self, name, default_unit=default_unit, parent=parent)

    def setValue(self, value):
        """

        :param float value: pixel size in centimeter
        """
        _value = value
        if type(_value) is str:
            try:
                _value = float(_value)
            except Exception as error:
                raise ValueError("Given string does not represent a float", error)
        _value = _value * metricsystem.centimeter
        MetricEntry.setValue(self, _value)

    def getValue(self):
        """

        :return: the value in meter
        :rtype: float
        """
        value = MetricEntry.getValue(self)
        return value / metricsystem.centimeter



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "11/01/2023"


import pathlib
from datetime import datetime
from urllib.parse import urlparse

from processview.core.dataset import Dataset, DatasetIdentifier
from tomoscan.esrf.identifier.rawidentifier import (
    RawVolumeIdentifier as _RawVolumeIdentifier,
)
from tomoscan.esrf.identifier.url_utils import split_path
from tomoscan.esrf.volume.rawvolume import RawVolume as _RawVolume

from tomwer.core.volume.volumebase import TomwerVolumeBase


class RawVolumeIdentifier(_RawVolumeIdentifier, DatasetIdentifier):
    """Identifier specific to .vol volume"""

    def __init__(self, object, file_path, metadata=None):
        super().__init__(object, file_path)
        DatasetIdentifier.__init__(self, RawVolume.from_identifier, metadata=metadata)

    @staticmethod
    def from_str(identifier):
        info = urlparse(identifier)
        paths = split_path(info.path)
        if len(paths) == 1:
            vol_file = paths[0]
            tomo_type = None
        elif len(paths) == 2:
            tomo_type, vol_file = paths
        else:
            raise ValueError("Failed to parse path string:", info.path)
        if tomo_type is not None and tomo_type != RawVolumeIdentifier.TOMO_TYPE:
            raise TypeError(
                f"provided identifier fits {tomo_type} and not {RawVolumeIdentifier.TOMO_TYPE}"
            )
        return RawVolumeIdentifier(object=RawVolume, file_path=vol_file)

    def long_description(self) -> str:
        """used for processview header tooltip for now"""
        return self.to_str()


class RawVolume(_RawVolume, TomwerVolumeBase, Dataset):
    @staticmethod
    def from_identifier(identifier):
        """Return the Dataset from a identifier"""
        if not isinstance(identifier, RawVolumeIdentifier):
            raise TypeError(
                f"identifier should be an instance of {RawVolumeIdentifier}"
            )
        return RawVolume(
            file_path=identifier.file_path,
        )

    def get_identifier(self) -> RawVolumeIdentifier:
        if self.url is None:
            raise ValueError("no file_path provided. Cannot provide an identifier")

        try:
            stat = pathlib.Path(self.url.file_path()).stat()
        except Exception:
            stat = None

        return RawVolumeIdentifier(
            object=self,
            file_path=self.url.file_path(),
            metadata={
                "name": self.url.file_path(),
                "creation_time": (
                    datetime.fromtimestamp(stat.st_ctime) if stat else None
                ),
                "modification_time": (
                    datetime.fromtimestamp(stat.st_ctime) if stat else None
                ),
            },
        )



import os
import numpy
from tomwer.core.process.control.emailnotifier import format_email_info, _ls_tomo_obj
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.volume.hdf5volume import HDF5Volume
from tomwer.core.utils.scanutils import HDF5MockContext


def test__ls_tomo_obj(tmp_path):
    """simple test of the `_ls_tomo_obj` function"""
    for i in range(2):
        open(os.path.join(tmp_path, f"{i}.nx"), "a").close()

    scan = NXtomoScan(
        scan=os.path.join(tmp_path, "1.nx"),
        entry="entry0000",
    )
    assert isinstance(scan, NXtomoScan)

    assert len(_ls_tomo_obj(scan)) == 2

    volume = HDF5Volume(
        file_path=os.path.join(tmp_path, "myvolume.hdf5"),
        data_path="myvolume",
        data=numpy.linspace(0, 10, 100 * 100 * 3).reshape(  # pylint: disable=E1121
            3, 100, 100
        ),
    )
    volume.save()
    assert len(_ls_tomo_obj(volume)) == 3


def test_format_email_info(tmp_path):
    """
    simple test of formatting some information related to an email
    """
    with HDF5MockContext(
        scan_path=os.path.join(tmp_path, "test", "scan"), n_proj=100
    ) as scan:
        res = format_email_info(
            my_str="{tomo_obj_short_id} \n {tomo_obj_id} \n {ls_tomo_obj} \n {timestamp} \n {footnote}",
            tomo_obj=scan,
        )

        for keyword in (
            "{tomo_obj_short_id}",
            "{tomo_obj_id}",
            "{ls_tomo_obj}",
            "{timestamp}",
            "{footnote}",
        ):
            assert keyword not in res



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "31/05/2017"


# TODO: change this for enum

OBSERVATION_STATUS = {
    "not processing": 0,  # data watcher is not doing anything
    "parsing": 1,  # data watcher is parsing folders
    "none found": 2,  # data watcher haven't found anything
    "starting": 3,  # data watcher have found an acquisition starting
    "started": 4,  # data watcher have found an acquisition started
    "waiting for acquisition ending": 5,  # data watcher is waiting for all .edf file to be copied
    "acquisition ended": 6,  # data watcher have found an acquisition completed
    "acquisition canceled": 7,  # data watcher have found an acquisition canceled
    "failure": -1,  # data watcher have encountered an issue. Should be associated with an info describing it
    "aborted": -2,  # if aborted by acquisition
}

DICT_OBS_STATUS = {}
for name, value in OBSERVATION_STATUS.items():
    DICT_OBS_STATUS[value] = name

" Possible status of a data watcher observation"

DET_END_XML = "[scan].xml"
"""DET_END for detection end.
In this case we are looking for the scan.xml file. On it is here then the
acquisition is considered ended
"""

PARSE_INFO_FILE = "[scan].info"
"""
In this case we will end for the .info to be here and then parse it to know how
many .edf file we are waiting for and wait for all of them to be recorded.
"""

DET_END_USER_ENTRY = "from file pattern"
"""DET_END for detection end.
In this case the user specify the pattern we are looking for
"""

DET_END_METHODS = (DET_END_XML, PARSE_INFO_FILE, DET_END_USER_ENTRY)
"""List of option that can notice the end of the acquisition"""


BLISS_SCAN_END = "bliss scan - end_time dataset"


NXtomo_END = "already converted NXtomo"

EDF_SCAN_END = "Any EDF scan"



from typing import Optional
from silx.gui import qt
from tomwer.gui import icons as tomwericons
from tomwer.core.utils.dictutils import concatenate_dict


class EmailComposition(qt.QWidget):
    """
    Widget used to compose an email

    Please see https://confluence.esrf.fr/display/SCKB/Rules+about+email for email rules at esrf
    """

    sigChanged = qt.Signal()
    """emit when the composition changed"""

    def __init__(self, parent: Optional[qt.QWidget]) -> None:
        super().__init__(parent)
        self.setLayout(qt.QFormLayout())
        # from
        self._fromAddressesQLE = qt.QLineEdit(self)
        singleEmailValidator = qt.QRegularExpressionValidator(
            qt.QRegularExpression(
                r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)"
            ),
            self._fromAddressesQLE,
        )
        self._fromAddressesQLE.setValidator(singleEmailValidator)
        self.layout().addRow("from", self._fromAddressesQLE)
        # to
        self._toAdressesQLE = qt.QLineEdit(self)
        severalEmailValidator = qt.QRegularExpressionValidator(
            qt.QRegularExpression(
                r"(([a-zA-Z0-9_\-\.]+)@((\[[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.)|(([a-zA-Z0-9\-]+\.)+))([a-zA-Z]{2,4}|[0-9]{1,3})(\]?)(\s*;\s*|\s*$))*"
            ),
            self,
        )
        self._toAdressesQLE.setValidator(severalEmailValidator)
        self.layout().addRow("to", self._toAdressesQLE)
        # subject
        self._subjectQLE = qt.QLineEdit("tomwer: {tomo_obj_short_id} received", self)
        tooltip = """
        this section can contain some predefine `keywords` and will be formatting before sending the email.\n
        Reserve keywords are: \n
        - {tomo_obj_short_id}: tomo_obj short identifier
        - {tomo_obj_id}: tomo_obj id
        - {ls_tomo_obj}: ls of the scan folder
        - {timestamp}: current time
        - {footnote}: some footnote defined by tomwer
        - {dataset_processing_states}: list dataset state regarding met processes
        """
        self._subjectQLE.setToolTip(tooltip)
        self.layout().addRow("subject", self._subjectQLE)
        # text
        self._textQLE = qt.QPlainTextEdit(
            "{tomo_obj_id} received at {timestamp} \n\nprocesses:\n{dataset_processing_states}\n\nfile listening:\n{ls_tomo_obj}\n\nInfo: {footnote}"
        )
        self._textQLE.setToolTip(tooltip)
        self.layout().addRow(self._textQLE)

        # connect signal / slot
        self._fromAddressesQLE.editingFinished.connect(self.sigChanged)
        self._toAdressesQLE.editingFinished.connect(self.sigChanged)
        self._subjectQLE.editingFinished.connect(self.sigChanged)
        self._textQLE.textChanged.connect(self.sigChanged)

    def getToAddrs(self) -> tuple:
        adresses = self._toAdressesQLE.text().replace(" ", "")
        adresses.replace(",", ";")
        return tuple(set(adresses.split(";")))

    def getConfiguration(self) -> dict:
        return {
            "subject": self._subjectQLE.text(),
            "from_addr": self._fromAddressesQLE.text(),
            "to_addrs": self.getToAddrs(),
            "text": self._textQLE.toPlainText(),
        }

    def setConfiguration(self, config: dict):
        assert isinstance(config, dict)
        subject = config.get("subject", None)
        if subject is not None:
            self._subjectQLE.setText(str(subject))
        from_addr = config.get("from_addr", None)
        if from_addr is not None:
            self._fromAddressesQLE.setText(str(from_addr))
        to_addrs = config.get("to_addrs", None)
        if to_addrs is not None:
            if not isinstance(to_addrs, str):
                to_addrs = ";".join(to_addrs)
            self._toAdressesQLE.setText(str(to_addrs))
        text = config.get("text", None)
        if text is not None:
            self._textQLE.setPlainText(str(text))


class EmailSettings(qt.QWidget):
    """
    Widget to set up mail server
    """

    sigChanged = qt.Signal()
    """emit when the settings changed"""

    def __init__(self, parent: Optional[qt.QWidget]) -> None:
        super().__init__(parent)
        self.setLayout(qt.QFormLayout(self))
        # server
        self._hostQLE = qt.QLineEdit("smtps.esrf.fr", self)
        self.layout().addRow("host", self._hostQLE)
        # port
        self._portQLE = qt.QLineEdit("0", self)
        self._portQLE.setValidator(qt.QIntValidator(self._portQLE))
        self.layout().addRow("port", self._portQLE)

        # connect signal / slot
        self._hostQLE.editingFinished.connect(self.sigChanged)
        self._portQLE.editingFinished.connect(self.sigChanged)

    def getConfiguration(self) -> dict:
        return {"host": self._hostQLE.text(), "port": int(self._portQLE.text())}

    def setConfiguration(self, config: dict):
        server = config.get("host", None)
        if server is not None:
            self._hostQLE.setText(str(server))
        port = config.get("port", None)
        if port is not None:
            self._portQLE.setText(str(port))


class Emailwidget(qt.QTabWidget):
    """
    Main widget to send email
    """

    sigChanged = qt.Signal()
    """emit when the email settings or composition changed"""

    def __init__(self, parent: Optional[qt.QWidget] = None) -> None:
        super().__init__(parent)

        self.setTabPosition(qt.QTabWidget.East)

        self._compositionWidget = EmailComposition(self)
        composeIcon = tomwericons.getQIcon("compose")
        self.addTab(
            self._compositionWidget,
            composeIcon,
            "",
        )

        self._serverSettingsWidget = EmailSettings(self)
        settingsIcon = tomwericons.getQIcon("parameters")
        self.addTab(
            self._serverSettingsWidget,
            settingsIcon,
            "",
        )

        # connect signal / slot
        self._compositionWidget.sigChanged.connect(self.sigChanged)
        self._serverSettingsWidget.sigChanged.connect(self.sigChanged)

    def getConfiguration(self) -> dict:
        return concatenate_dict(
            self._serverSettingsWidget.getConfiguration(),
            self._compositionWidget.getConfiguration(),
        )

    def setConfiguration(self, config: dict):
        self._serverSettingsWidget.setConfiguration(config)
        self._compositionWidget.setConfiguration(config)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "05/04/2019"


import gc
import os
import shutil
import tempfile
import time
import unittest

import h5py
import numpy
import pytest
from nxtomomill.converter import from_h5_to_nx
from nxtomomill.io.config.hdf5config import TomoHDF5Config
from nxtomo.nxobject.nxdetector import ImageKey

from silx.io.utils import open as open_hdf5
from tomoscan.io import HDF5File
from tomoscan.validator import is_valid_for_reconstruction

from tomwer.core.process.control.datalistener import DataListener
from tomwer.core.process.control.scantransfer import ScanTransferTask
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.utils.scanutils import MockEDF
from tomwer.synctools.rsyncmanager import RSyncManager
from tomwer.tests.datasets import TomwerCIDatasets


class TestDataTransferIO(unittest.TestCase):
    """Test inputs and outputs types of the handler functions"""

    def setUp(self):
        super().setUp()
        self._used_folders = []
        self.origin_folder = None
        self.output_folder = None

    def tearDown(self):
        for folder in self._used_folders:
            shutil.rmtree(folder)
        super().tearDown()

    def testInputOutput(self):
        """Test that io using TomoBase instance work"""
        for input_type in (dict, TomwerScanBase):
            for serialize_output_data in (True, False):
                with self.subTest(
                    return_dict=serialize_output_data,
                    input_type=input_type,
                ):
                    output_folder = tempfile.mkdtemp()
                    origin_folder = tempfile.mkdtemp()
                    scan_folder = os.path.join(origin_folder, "scan_toto")
                    os.makedirs(scan_folder)
                    os.makedirs(os.path.join(output_folder, "scan_toto"))
                    self._used_folders.append(output_folder)
                    self._used_folders.append(origin_folder)

                    scan = MockEDF.mockScan(
                        scanID=scan_folder,
                        nRadio=10,
                        nRecons=1,
                        nPagRecons=4,
                        dim=10,
                    )

                    transfert_process = ScanTransferTask(
                        inputs={
                            "data": scan,
                            "dest_dir": output_folder,
                            "serialize_output_data": serialize_output_data,
                        }
                    )

                    self.assertTrue(os.path.exists(output_folder))
                    input_obj = scan
                    if input_obj is dict:
                        input_obj = input_obj.to_dict()
                    transfert_process.run()
                    out = transfert_process.outputs.data
                    if serialize_output_data:
                        self.assertTrue(isinstance(out, dict))
                    else:
                        self.assertTrue(isinstance(out, TomwerScanBase))


@pytest.mark.skipif(not RSyncManager().has_rsync(), reason="requires rsync")
class TestBlissDataTransfer(unittest.TestCase):
    """Make sure we can transfer data from bliss acquisition"""

    def setUp(self):
        self.input_dir = tempfile.mkdtemp()
        self.output_dir = tempfile.mkdtemp()
        shutil.copytree(
            TomwerCIDatasets.get_dataset("bliss/sample"),
            os.path.join(self.input_dir, "sample"),
        )

        self._proposal_file = os.path.join(
            self.input_dir, "sample", "ihpayno_sample.h5"
        )
        self._sample_file = os.path.join(
            self.input_dir, "sample", "sample_29042021", "sample_29042021.h5"
        )
        self._sample_file_entry = "1.1"
        assert os.path.exists(self._sample_file)

        # mock data listener: the processing of the Data transfer requires
        # knowledge of bliss files origin.
        self.assertTrue(os.path.exists(self._proposal_file))
        data_listener = DataListener()
        scans = data_listener.process_sample_file(
            sample_file=self._sample_file,
            entry=self._sample_file_entry,
            proposal_file=self._proposal_file,
            master_sample_file=None,
        )
        self.scan = scans[0]
        self.assertTrue(os.path.exists(self.scan.process_file))

    def tearDown(self):
        shutil.rmtree(self.input_dir)
        shutil.rmtree(self.output_dir)
        gc.collect()

    def testDataTransfer(self):
        """Make sure the data transfer is able to retrieve the scan,
        proposal file and scan file to transfer.
        Check that only the specific scan folders will be copy and removed
        and the other won't be affected.
        """
        out_proposal = os.path.join(self.output_dir, "ihpayno_sample.h5")
        self.assertFalse(os.path.exists(out_proposal))
        out_sample_file = os.path.join(
            self.output_dir, "sample_29042021", "sample_29042021.h5"
        )
        self.assertFalse(os.path.exists(out_sample_file))
        out_included_scans = [
            os.path.join(self.output_dir, "sample_29042021", "scan0002")
        ]
        for scan_path in out_included_scans:
            self.assertFalse(os.path.exists(scan_path))
        out_not_included_scans = [
            os.path.join(self.output_dir, "sample_29042021", "scan0004"),
            os.path.join(self.output_dir, "sample_29042021", "scan0006"),
            os.path.join(self.output_dir, "sample_29042021", "scan0008"),
        ]
        for scan_path in out_not_included_scans:
            self.assertFalse(os.path.exists(scan_path))

        process = ScanTransferTask(
            inputs={
                "data": self.scan,
                "block": True,
                "dest_dir": self.output_dir,
            }
        )
        process.run()

        time.sleep(1)
        self.assertTrue(os.path.exists(out_proposal), f"{out_proposal} does not exists")
        self.assertTrue(
            os.path.exists(out_sample_file), f"{out_sample_file} does not exists"
        )
        for scan_path in out_included_scans:
            self.assertTrue(os.path.exists(scan_path))
        for scan_path in out_not_included_scans:
            self.assertFalse(os.path.exists(scan_path))

        # ensure we can convert it using nxtomomill once transferred
        output_file = out_sample_file.replace("29042021.h5", "29042021.nx")
        assert output_file != out_sample_file
        configuration = TomoHDF5Config()
        configuration.input_file = out_sample_file
        configuration.output_file = output_file
        configuration.single_file = True
        configuration.file_extension = ".nx"
        from_h5_to_nx(configuration, input_callback=lambda *arg: 0)
        assert os.path.exists(output_file)
        assert is_valid_for_reconstruction(NXtomoScan(output_file, entry="entry0000"))


@pytest.mark.skipif(not RSyncManager().has_rsync(), reason="requires rsync")
class NXTomoDataTransferBase(unittest.TestCase):
    """Make sure we can transfer data from an NXTomo with linked dependancies"""

    def setUp(self) -> None:
        super().setUp()
        self._src_dir = tempfile.mkdtemp()
        self._dst_dir = tempfile.mkdtemp()

    def create_nexus_file(self, relative_links=True):
        dims = 100, 200
        n_darks = 2
        n_flats = 3
        n_projections = 10
        scans = {
            "scan0002": (ImageKey.DARK_FIELD, n_darks),
            "scan0003": (ImageKey.FLAT_FIELD, n_flats),
            "scan0004": (ImageKey.PROJECTION, n_projections),
        }
        self.nexus_file_path = os.path.join(self._src_dir, "my_acquisition.nx")
        sources = []
        sources_len = []
        image_keys = []
        # create raw data
        for scan_name, (image_key, n_frames) in scans.items():
            dir_path = os.path.join(self._src_dir, scan_name)
            os.makedirs(dir_path)
            file_path = os.path.join(dir_path, "pcolinux_0000.h5")
            with HDF5File(file_path, mode="w") as h5s:
                pcolinux_grp = h5s.require_group("entry0000/instrument/pcolinux")
                shape = (n_frames, dims[1], dims[0])
                data = numpy.random.random(shape)
                data *= 100
                pcolinux_grp["data"] = data.astype(dtype=numpy.uint16)
                # create a link tp measurement
                h5s["entry0000/measurement/data"] = h5py.SoftLink(
                    "/".join(("/entry0000", "instrument", "pcolinux", "data"))
                )
            vsource_path = os.path.join(scan_name, "pcolinux_0000.h5")
            if not relative_links:
                vsource_path = os.path.normpath(
                    os.path.join(self._src_dir, vsource_path)
                )
            sources.append(
                h5py.VirtualSource(
                    vsource_path,
                    name="entry0000/instrument/pcolinux/data",
                    shape=shape,
                )
            )
            sources_len.append(n_frames)
            image_keys.extend([image_key.value] * n_frames)
        # create virtual dataset to the raw data
        with HDF5File(self.nexus_file_path, mode="w") as h5s:
            entry_node = h5s.require_group("entry0000")

            # write detector
            detector_node = entry_node.require_group("detector")
            shape = (
                n_darks + n_flats + n_projections,
                dims[1],
                dims[0],
            )
            layout = h5py.VirtualLayout(shape=shape, dtype=numpy.uint16)
            i_frame = 0
            for v_source, v_source_length in zip(sources, sources_len):
                layout[i_frame : i_frame + v_source_length] = v_source
                i_frame += v_source_length
            n_frames = i_frame
            detector_node.create_virtual_dataset("data", layout, fillvalue=numpy.nan)
            detector_node["count_time"] = numpy.array([0.1] * n_frames)
            detector_node["distance"] = 0.17
            detector_node["distance"].attrs["unit"] = "m"
            # write general information
            entry_node["definition"] = "NXtomo"
            entry_node["end_time"] = "2021-03-03T15:51:00.986554+01:00"
            entry_node["start_time"] = "2021-03-03T15:43:00.693382+01:00"
            entry_node["title"] = "0001"
            entry_node["field_of_view"] = "Half"
            entry_node["tomo_n"] = n_frames
            entry_node["x_pixel_size"] = 3.25e-7
            entry_node["y_pixel_size"] = 3.25e-7

            # write beam: add an absolute link for energy to ensure this is style valid and that relative files won't be moved
            file_with_energy = os.path.join(self._src_dir, "energy.h5")
            with HDF5File(file_with_energy, mode="w") as energy_file:
                energy_file["energy"] = 19
                energy_file["energy"].attrs["unit"] = "keV"

            entry_node["instrument/beam/incident_energy"] = h5py.ExternalLink(
                file_with_energy, "/energy"
            )

            entry_node["beam"] = h5py.SoftLink(
                "/".join(("/entry0000", "instrument", "beam"))
            )
            # write sample
            sample_node = entry_node.require_group("sample")
            sample_node["name"] = "sample name"
            sample_node["image_key"] = numpy.array(image_keys)
            sample_node["image_key_control"] = numpy.array(image_keys)
            sample_node["rotation_angle"] = numpy.linspace(0, 360, n_frames)
            sample_node["x_translation"] = numpy.array([0.13] * n_frames)
            sample_node["y_translation"] = numpy.array([0.0063] * n_frames)
            sample_node["z_translation"] = numpy.array([0.6] * n_frames)

    def tearDown(self) -> None:
        shutil.rmtree(self._src_dir)
        shutil.rmtree(self._dst_dir)
        return super().tearDown()

    def _test(self):
        src_scan = NXtomoScan(self.nexus_file_path, entry="entry0000")
        assert is_valid_for_reconstruction(src_scan)
        process = ScanTransferTask(
            inputs={
                "data": src_scan,
                "turn_off_print": True,
                "dest_dir": self._dst_dir,
                "block": True,
                "noRsync": False,
            }
        )
        process.run()
        dst_scan = process.outputs.data
        assert dst_scan.master_file != self.nexus_file_path
        assert is_valid_for_reconstruction(dst_scan, check_values=True)
        with open_hdf5(src_scan.master_file) as h5s_src:
            src_dataset = h5s_src["entry0000/detector/data"][...]
        with open_hdf5(dst_scan.master_file) as h5s_dst:
            dst_dataset = h5s_dst["entry0000/detector/data"][...]
        assert numpy.array_equal(src_dataset, dst_dataset)


@pytest.mark.skipif(not RSyncManager().has_rsync(), reason="requires rsync")
class TestNXtomoTransferRelativeLink(NXTomoDataTransferBase):
    """Ensure NXtomo conversion will work if it contains relative links"""

    def setUp(self) -> None:
        super().setUp()
        self.create_nexus_file(relative_links=True)

    def test(self):
        super()._test()
        # make sure there is the .nx file there and the folder containing file linked the relative
        assert (
            len(
                os.listdir(os.path.join(self._dst_dir, self._src_dir.split(os.sep)[-1]))
            )
            == 4
        )


@pytest.mark.skipif(not RSyncManager().has_rsync(), reason="requires rsync")
class TestNXtomoTransferAbsoluteLink(NXTomoDataTransferBase):
    """Ensure NXtomo conversion will work if it contains absolute links"""

    def setUp(self) -> None:
        super().setUp()
        self.create_nexus_file(relative_links=False)

    def test(self):
        super()._test()
        # make sure there is only the .nx file there
        assert (
            len(
                os.listdir(os.path.join(self._dst_dir, self._src_dir.split(os.sep)[-1]))
            )
            == 1
        )



from .utils import format_stderr_stdout  # noqa F401
from .utils import get_default_directory  # noqa F401
from .utils import get_linked_files_with_entry  # noqa F401
from .utils import get_linked_files_with_vds  # noqa F401
from .utils import get_slice_data  # noqa F401



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "16/03/2020"


import logging
import time
from collections import namedtuple
from datetime import datetime
from multiprocessing import Lock, Manager

from silx.gui import qt

from tomwer.core.process.control.datalistener.rpcserver import (
    BlissAcquisition,
    TangoAcquisitionStatus,
    _BaseDataListenerThread,
)
from tomwer.utils import docstring

_logger = logging.getLogger(__name__)


class DataListenerQThread(_BaseDataListenerThread, qt.QThread):
    """Implementation of the _BaseDataListenerThread with a QThread.

    As jsonrpc is using gevent we want to limit the 'crossover' of gevent and
    qt. So the server is launch in another process and bot processes are
    sharing list of acquisitions (added, ended...).
    """

    sigAcquisitionStarted = qt.Signal(tuple)
    """Signal emitted when an acquisition is started. Tuple is:
    (master_file, master_entry)"""
    sigAcquisitionEnded = qt.Signal(tuple)
    """Signal emitted when an acquisition is ended. Tuple is
    (master_file, master_entry, succeed)"""
    sigScanAdded = qt.Signal(tuple)
    """Signal emitted when a scan is added to an acquisition. Tuple is
    (master_file, master_entry, scan_entry)"""
    sigServerStop = qt.Signal()
    """Signal if the rpc-server have been turn off"""

    _WAIT_BTW_COLLECT = 0.5

    def __init__(self, host, port, acquisitions):
        qt.QThread.__init__(self)
        _BaseDataListenerThread.__init__(
            self, host=host, port=port, acquisitions=acquisitions
        )
        self._manager = Manager()

        self._lock = Lock()
        self._sequence_started = self._manager.list()
        self._sequence_ended = self._manager.list()
        self._scan_added = self._manager.list()
        self._stop = False

    @docstring(_BaseDataListenerThread.sequence_started)
    def sequence_started(self, acquisition: BlissAcquisition) -> None:
        assert isinstance(acquisition, BlissAcquisition)
        with self._lock:
            self._sequence_started.append(
                (
                    acquisition.master_file,
                    acquisition.entry,
                    acquisition.proposal_file,
                    acquisition.sample_file,
                )
            )

    def _collect_sequence_started(self):
        with self._lock:
            for scan in self._sequence_started:
                master_file, entry, proposal_file, sample_file = scan
                self.sigAcquisitionStarted.emit(
                    (master_file, entry, proposal_file, sample_file)
                )

            self._sequence_started[:] = []

    @docstring(_BaseDataListenerThread.scan_added)
    def scan_added(self, acquisition: BlissAcquisition, scan_number: int) -> None:
        assert isinstance(acquisition, BlissAcquisition)
        with self._lock:
            self._scan_added.append(
                (
                    acquisition.master_file,
                    acquisition.entry,
                    acquisition.proposal_file,
                    acquisition.sample_file,
                    scan_number,
                )
            )

    def _collect_scan_added(self):
        with self._lock:
            for scan in self._scan_added:
                master_file, entry, proposal_file, sample_file, scan_number = scan
                self.sigScanAdded.emit(
                    (master_file, entry, proposal_file, sample_file, scan_number)
                )
            self._scan_added[:] = []

    @docstring(_BaseDataListenerThread.sequence_ended)
    def sequence_ended(self, acquisition: BlissAcquisition) -> None:
        assert isinstance(acquisition, BlissAcquisition)
        with self._lock:
            self._sequence_ended.append(
                (
                    acquisition.master_file,
                    acquisition.entry,
                    acquisition.proposal_file,
                    acquisition.sample_file,
                    not acquisition.has_error,
                )
            )

    def _collect_sequence_ended(self):
        with self._lock:
            for scan in self._sequence_ended:
                master_file, entry, proposal_file, sample_file, success = scan
                self.sigAcquisitionEnded.emit(
                    (master_file, entry, proposal_file, sample_file, success)
                )
            self._sequence_ended[:] = []

    @docstring(_BaseDataListenerThread.sequence_failed)
    def sequence_failed(self, acquisition: BlissAcquisition):
        assert isinstance(acquisition, BlissAcquisition)
        with self._lock:
            self._sequence_ended.append(
                (
                    acquisition.master_file,
                    acquisition.entry,
                    acquisition.proposal_file,
                    acquisition.sample_file,
                    False,
                )
            )

    def start(self):
        _BaseDataListenerThread.start(self)
        qt.QThread.start(self)

    def join(self, timeout=None):
        if timeout is None:
            self.wait()
        else:
            self.wait(timeout)

    def stop(self):
        _BaseDataListenerThread.stop(self)
        self._stop = True
        self.join()

    def _make_sure_server_still_running(self):
        if not self._server.is_alive():
            self.sigServerStop.emit()

    def run(self):
        while not self._stop:
            try:
                self._collect_sequence_started()
                self._collect_scan_added()
                self._collect_sequence_ended()
                self._make_sure_server_still_running()
            except Exception as e:
                _logger.error(e)
                break
            else:
                time.sleep(DataListenerQThread._WAIT_BTW_COLLECT)


_mock_acquisition_info = namedtuple(
    "_mock_acquisition_info", ["master_file", "entry", "scan_numbers", "waiting_time"]
)


class MockDataListenerQThread(DataListenerQThread):
    """
    Overwrite run() function and mock an acquisition for CI avoiding to have a
    tango server and tango install.
    """

    def __init__(self, host, port, acquisitions: list, mock_acquisitions: list):
        DataListenerQThread.__init__(
            self, host=host, port=port, acquisitions=acquisitions
        )
        assert isinstance(mock_acquisitions, (list, tuple))
        for acqui in mock_acquisitions:
            assert type(acqui) is _mock_acquisition_info
        self.mock_acquisitions = mock_acquisitions
        "waiting time between each scan / step of the acquisition"

    def get_device(self, device_name):
        return None

    def run(self) -> None:
        _logger.info("mock an acquisition using tango / tango")

        for mock_acqui in self.mock_acquisitions:
            if self._stop:
                return

            # first scan is the definition of the acquisition
            now = datetime.now()
            current_acquisition = BlissAcquisition(
                file_path=mock_acqui.master_file,
                entry_name=mock_acqui.entry,
                start_time=now.strftime("%H:%M:%S"),
                proposal_file=None,
                sample_file=None,
            )
            self.acquisitions.append(current_acquisition)
            self.sequence_started(current_acquisition)

            time.sleep(mock_acqui.waiting_time)

            # add scan numbers if any
            for scan_number in mock_acqui.scan_numbers:
                current_acquisition.add_scan_number(scan_number)
                current_acquisition.set_status(TangoAcquisitionStatus.ON_GOING)
                self.scan_added(
                    acquisition=current_acquisition, scan_number=scan_number
                )
                if self._stop:
                    return
                time.sleep(mock_acqui.waiting_time)

            # end acquisition
            now = datetime.now()
            current_acquisition.end(
                end_time=now.strftime("%H:%M:%S"), succeed=True, error=None
            )
            self._rpc_sequence_ended(
                saving_file=current_acquisition.sample_file,
                sequence_scan_number=current_acquisition.entry,
                success=True,
            )
            time.sleep(mock_acqui.waiting_time)

        # wait until stop the thread
        while not self._stop:
            time.sleep(0.1)



from typing import Optional
import h5py
from silx.gui import qt
from silx.io.utils import open as open_hdf5

from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.gui import icons


class NXtomoProxyWarmer(qt.QWidget):
    """
    Widget to warm in the case the NXtomo entry is a SoftLink or an ExternalLink (so editing it will modify another file).
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setLayout(qt.QHBoxLayout())
        # icon
        # warning about requires parameters from a nabu slice reconstruction
        self._warningIcon = qt.QLabel(
            "It seems this entry points to a soft link. Be carreful. Editing metadata will affect another file",
            parent=self,
        )
        self._warningIcon.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
        self._warningIcon.setAlignment(qt.Qt.AlignRight | qt.Qt.AlignVCenter)
        warning_icon = icons.getQIcon("warning")
        self._warningIcon.setPixmap(warning_icon.pixmap(20, 20))
        self.layout().addWidget(self._warningIcon)
        # label
        self._warningLabel = qt.QLabel(
            "It seems the NXtomo you are editing is pointing to another dataset.",
            parent=self,
        )
        self._warningLabel.setAlignment(qt.Qt.AlignRight | qt.Qt.AlignVCenter)
        self._warningLabel.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
        font = self._warningLabel.font()
        font.setBold(True)
        font.setPixelSize(10)
        self._warningLabel.setFont(font)
        self.layout().addWidget(self._warningLabel)
        # set up
        self.setScan(None)

    def setScan(self, scan: Optional[NXtomoScan]):
        if scan is None:
            self._activateWarning(False)
        elif isinstance(scan, NXtomoScan):
            with open_hdf5(scan.master_file) as h5f:
                entry = h5f.get(
                    name=scan.entry, getclass=True, getlink=True, default=None
                )
                self._activateWarning(entry in (h5py.ExternalLink, h5py.SoftLink))
        else:
            raise TypeError(f"{scan} is expected to be an instance of {NXtomoScan}")

    def _activateWarning(self, activate: bool):
        self._warningIcon.setVisible(activate)
        self._warningLabel.setVisible(activate)



import logging

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output
from orangecontrib.tomwer.orange.managedprocess import TomwerWithStackStack
from silx.gui import qt
from typing import Optional

from tomwer.core.futureobject import FutureTomwerObject
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.reconstruction.nabu.helical import HelicalPrepareWeightsDouble
from tomwer.core.process.reconstruction.nabu.helical import (
    NabuHelicalPrepareWeightsDouble,
)
from tomwer.core.scan.nxtomoscan import NXtomoScan

_logger = logging.getLogger(__name__)


class NabuHelicalPrepareWeightsDoubleOW(
    TomwerWithStackStack,
    ewokstaskclass=NabuHelicalPrepareWeightsDouble,
):
    """
    widget used to call the `nabu-helical-prepare-weights-double` application on a dedicated thread. It define weights map and double flat field.

    :param parent: the parent widget
    """

    # note of this widget should be the one registered on the documentation
    name = "helical prerate weights double"
    id = "orangecontrib.tomwer.widgets.reconstruction.NabuHelicalPrepareWeightsDoubleOW.NabuHelicalPrepareWeightsDoubleOW"
    description = "compute map of weights requested for nabu helical reconstruction"
    icon = "icons/nabu_prepare_weights_double.svg"
    priority = 199
    keywords = [
        "tomography",
        "nabu",
        "reconstruction",
        "nabu-helical",
        "helical",
        "weights",
        "prepare",
    ]

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    _ewoks_default_inputs = Setting(
        {
            "data": None,
            "transition_width_vertical": 50,
            "transition_width_horizontal": 50,
            "processes_file": "",
            "rotation_axis_position": 0,
        }
    )

    _ewoks_inputs_to_hide_from_orange = (
        "progress",
        "processes_file",
        "transition_width_vertical",
        "transition_width_horizontal",
        "rotation_axis_position",
    )

    sigScanReady = qt.Signal(TomwerScanBase)
    "Signal emitted when a scan is ended"

    TIMEOUT = 30

    class Inputs:
        data = Input(
            name="data",
            type=TomwerScanBase,
            doc="one scan to be process",
            default=True,
            multiple=False,
        )

    class Outputs:
        data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")
        future_tomo_obj = Output(
            name="future_tomo_obj",
            type=FutureTomwerObject,
            doc="future object (process remotely)",
        )

    LOGGER = _logger

    def __init__(self, parent=None):
        super().__init__(parent)
        self.__scan = None

        # gui definition
        _layout = gui.vBox(self.mainArea, self.name).layout()
        self.widget = HelicalPrepareWeightsDouble(parent=self)
        _layout.addWidget(self.widget)

        ## connect signal / slot
        self.widget.sigConfigChanged.connect(self._updateSettings)

        if isinstance(self.task_output_changed_callbacks, set):
            self.task_output_changed_callbacks.add(self._notify_state)
        elif isinstance(self.task_output_changed_callbacks, list):
            self.task_output_changed_callbacks.append(self._notify_state)
        else:
            raise NotImplementedError

        ## handle settings
        self._loadSettings()
        self.task_executor_queue.sigComputationStarted.connect(self._newTaskStarted)

    def _updateSettings(self):
        config = self.widget.getConfiguration()
        for key in ("transition_width", "processes_file"):
            self._ewoks_default_inputs[key] = config[key]  # pylint: disable=E1137

    @property
    def request_input(self):
        return self.__request_input

    @request_input.setter
    def request_input(self, request):
        self.__request_input = request

    def get_task_inputs(self):
        assert self.__scan is not None
        return {
            "data": self.__scan,
            "transition_width": self.widget.getConfiguration()["transition_width"],
            "processes_file": self.widget.getConfiguration()["processes_file"],
        }

    def handleNewSignals(self) -> None:
        """Invoked by the workflow signal propagation manager after all
        signals handlers have been called.
        """
        # for now we want to avoid propagation any processing.
        # task will be executed only when the user validates the dialog
        data = super().get_task_inputs().get("data", None)
        if data is not None:
            if not isinstance(data, NXtomoScan):
                raise TypeError(
                    f"data is expected to be an instance of NXtomoScan. {type(data)} are not handled"
                )
            self.add(data.path)

    def _loadSettings(self):
        self.widget.setConfiguration(self._ewoks_default_inputs)

    def _newTaskStarted(self):
        try:
            task_executor = self.sender()
            scan = task_executor.current_task.inputs.data
            self.notify_on_going(scan)
        except Exception:
            pass

    def _notify_state(self):
        try:
            task_executor = self.sender()
            task_suceeded = task_executor.succeeded
            scan = task_executor.current_task.outputs.data
            if task_suceeded:
                self.notify_succeed(scan=scan)
            else:
                self.notify_failed(scan=scan)
        except Exception as e:
            _logger.error(f"failed to handle task finished callback. Raiseon is {e}")

    @Inputs.data
    def process_data(self, scan: Optional[TomwerScanBase]):
        if scan is None:
            return
        else:
            self.__scan = scan
            self.notify_pending(scan=scan)
            self.execute_ewoks_task()

    def sizeHint(self) -> qt.QSize:
        return qt.QSize(650, 60)



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/


__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "27/02/2019"


import glob
import logging
import os
from typing import Union
from urllib.parse import urlparse

import h5py
from tomoscan.esrf.identifier.url_utils import split_path
from tomoscan.factory import Factory as _oScanFactory
from tomoscan.identifier import BaseIdentifier, ScanIdentifier, VolumeIdentifier
from tomoscan.scanbase import TomoScanBase
from tomoscan.tomoobject import TomoObject
from nxtomo.application.nxtomo import NXtomo

from tomwer.utils import docstring

from .blissscan import BlissScan
from .edfscan import EDFTomoScan, EDFTomoScanIdentifier
from .hdf5scan import NXtomoScan, NXtomoScanIdentifier

_logger = logging.getLogger(__name__)


@docstring(_oScanFactory)
class ScanFactory(object):
    @docstring(_oScanFactory.create_scan_object)
    @staticmethod
    def create_tomo_object_from_identifier(
        identifier: Union[str, BaseIdentifier]
    ) -> TomoObject:
        """
        Create an instance of TomoScanBase from his identifier if possible

        :param str identifier: identifier of the TomoScanBase
        :raises: TypeError if identifier is not a str
        :raises: ValueError if identifier cannot be converted back to an instance of TomoScanBase
        """
        from tomoscan.identifier import BaseIdentifier as _TomoScanIdentifier

        if not isinstance(identifier, (str, BaseIdentifier, _TomoScanIdentifier)):
            raise TypeError(
                f"identifier is expected to be a str or an instance of {BaseIdentifier} not {type(identifier)}"
            )

        # step 1: convert identifier to an instance of BaseIdentifier if necessary
        if isinstance(identifier, str):
            info = urlparse(identifier)
            paths = split_path(info.path)
            scheme = info.scheme
            if len(paths) == 1:
                # insure backward compatibility. Originally (until 0.8) there was only one type which was scan
                tomo_type = ScanIdentifier.TOMO_TYPE
            elif len(paths) == 2:
                tomo_type, _ = paths
            else:
                raise ValueError("Failed to parse path string:", info.path)

            if tomo_type == VolumeIdentifier.TOMO_TYPE:
                raise ValueError("This factory creates scan object and not volumes")

            elif tomo_type == ScanIdentifier.TOMO_TYPE:
                # otherwise consider this is a scan. Insure backward compatibility
                if scheme == "edf":
                    identifier = EDFTomoScanIdentifier.from_str(identifier=identifier)
                elif scheme == "hdf5":
                    identifier = NXtomoScanIdentifier.from_str(identifier=identifier)
                else:
                    raise ValueError(f"Scheme {scheme} not recognized")
            else:
                raise ValueError(f"{tomo_type} is not an handled tomo type")

        # step 2: convert identifier to a TomoBaseObject
        assert isinstance(identifier, BaseIdentifier)
        scheme = identifier.scheme
        tomo_type = identifier.tomo_type

        if scheme == "edf":
            if tomo_type == ScanIdentifier.TOMO_TYPE:
                return EDFTomoScan.from_identifier(identifier=identifier)
            else:
                raise NotImplementedError()
        elif scheme == "hdf5":
            if tomo_type == ScanIdentifier.TOMO_TYPE:
                return NXtomoScan.from_identifier(identifier=identifier)
            else:
                raise NotImplementedError()
        else:
            raise ValueError(f"Scheme {scheme} not recognized")

    @staticmethod
    def create_scan_object(scan_path, entry=None, accept_bliss_scan=False):
        """

        :param TextIOWrapper scan_path: path to the scan directory or file
        :param entry: entry on the file. Requested for hdf5 files
        :param accept_bliss_scan: if True the factory can return some BlissScan
                                  But this is only compatible with the
                                  Tomomill processing.
        :return: TomwerScanBase instance fitting the scan folder or scan path
        :rtype: tomwer.core.scan.scanbase.TomwerScanBase
        :raises: ValueError if scan_path is not containing a scan
        """
        if scan_path is None:
            raise ValueError("'scan_path' should be provided")
        if entry is not None and not entry.startswith("/"):
            entry = "/" + entry
        if os.path.isfile(scan_path) and ScanFactory.is_hdf5_tomo(scan_path):
            valid_entries = NXtomo.get_valid_entries(scan_path)
            if entry is None:
                if len(valid_entries) > 1:
                    _logger.warning(
                        "more than one entry found for %s."
                        "Pick the last entry" % scan_path
                    )
                entry = valid_entries[-1]
            elif entry not in valid_entries:
                raise ValueError(
                    f"entry {entry} is invalid. Does it exists ? Is the "
                    f"file NXTomo compliant ?. Valid entry are {valid_entries}"
                )
            return NXtomoScan(scan=scan_path, entry=entry)
        elif ScanFactory.is_edf_tomo(scan_path):
            return EDFTomoScan(scan=scan_path)
        elif accept_bliss_scan and BlissScan.is_bliss_file(scan_path):
            if entry is None:
                valid_entries = BlissScan.get_valid_entries(scan_path)
                if len(valid_entries) > 1:
                    _logger.warning(
                        f"more than one entry found for {scan_path}. Pick the last entry"
                    )
                entry = valid_entries[-1]
            return BlissScan(master_file=scan_path, entry=entry, proposal_file=None)

        raise ValueError(f"Unable to generate a scan object from {scan_path}")

    @docstring(_oScanFactory.create_scan_objects)
    @staticmethod
    def create_scan_objects(scan_path, accept_bliss_scan=True) -> tuple:
        scan_path = scan_path.rstrip(os.path.sep)
        if EDFTomoScan.is_tomoscan_dir(scan_path):
            return (EDFTomoScan(scan=scan_path),)
        elif NXtomoScan.is_tomoscan_dir(scan_path):
            scans = []
            master_file = NXtomoScan.get_master_file(scan_path=scan_path)
            entries = NXtomo.get_valid_entries(master_file)
            for entry in entries:
                scans.append(NXtomoScan(scan=scan_path, entry=entry, index=None))
            return tuple(scans)
        elif accept_bliss_scan and BlissScan.is_bliss_file(scan_path):
            scans = []
            for entry in BlissScan.get_valid_entries(scan_path):
                scans.append(
                    BlissScan(master_file=scan_path, entry=entry, proposal_file=None)
                )
            return tuple(scans)
        return tuple()

    @staticmethod
    def mock_scan(type_="edf"):
        """Mock a scan structure which is not associated to any real acquistion"""
        if type_ == "edf":
            return EDFTomoScan(scan=None)
        else:
            raise NotImplementedError("Other TomoScan are not defined yet")

    @staticmethod
    def create_scan_object_frm_dict(_dict):
        if TomoScanBase.DICT_TYPE_KEY not in _dict:
            raise ValueError(
                f"given dict is not recognized. Cannot find {TomoScanBase.DICT_TYPE_KEY}"
            )
        elif _dict[TomoScanBase.DICT_TYPE_KEY] == EDFTomoScan._TYPE:
            return EDFTomoScan(scan=None).load_from_dict(_dict)
        elif _dict[TomoScanBase.DICT_TYPE_KEY] == NXtomoScan._TYPE:
            return NXtomoScan.from_dict(_dict)
        else:
            raise ValueError(
                f"Scan type {_dict[TomoScanBase.DICT_TYPE_KEY]} is not managed"
            )

    @staticmethod
    def is_tomo_scandir(scan_path):
        """

        :param str scan_path: path to the scan directory or file
        :return: True if the given path / file is a tomo_scandir. For now yes by
                 default
        :rtype: bool
        """
        return True

    @staticmethod
    def is_edf_tomo(scan_path):
        """

        :param str scan_path: path to the scan directory or file
        :return: True if given path define a tomo scan based on .edf file
        :rtype: bool
        """
        if scan_path and os.path.isdir(scan_path):
            file_basename = os.path.basename(scan_path)
            has_info_file = (
                len(glob.glob(os.path.join(scan_path, file_basename + "*.info"))) > 0
            )
            not_lbs_scan_path = scan_path.replace("lbsram", "", 1)
            has_notlbsram_info_file = (
                len(
                    glob.glob(os.path.join(not_lbs_scan_path, file_basename + "*.info"))
                )
                > 0
            )
            if has_info_file or has_notlbsram_info_file:
                return True
        return False

    @staticmethod
    def is_hdf5_tomo(scan_path):
        """

        :param scan_path:
        :return:
        """
        if os.path.isfile(scan_path):
            return h5py.is_hdf5(scan_path)
        else:
            return NXtomoScan.directory_contains_scan(scan_path)



"""utils around screenshots.
"""

from .icatbase import IcatProcessedDataBase
from tomwer.core.scan.scanbase import TomwerScanBase


class IcatScreenshots(IcatProcessedDataBase):
    """Class to define an icat process publishing screenshot (so to the galery)"""

    def __init__(self, data_dir: str, screenshots: dict, scan: TomwerScanBase) -> None:
        super().__init__(data_dir, scan)
        if not isinstance(screenshots, dict):
            raise TypeError(
                f"Screenshots is expected to be a dict. Get {type(screenshots)} instead."
            )

        self._screenshots = screenshots

    @property
    def screenshots(self):
        return self._screenshots

    def to_dict(self):
        res = super().to_dict()
        res["screenshots_as_dict"] = self.screenshots
        return res
